"""This module handles the simulation of the serial connection in a
:class:`~turboctl.virtualpump.virtualpump.VirtualPump`.
"""

import os
import pty
import select
import threading
import time


class VirtualConnection():
    """A virtual serial connection.

    Data can be sent through a :class:`VirtualConnection` object by
    accessing the :attr:`user_end` and :attr:`virtual_end` attributes.
    The :attr:`port` property is a device name which can be given to
    the initializer of :class:`serial.Serial` as an argument.

    A :class:`VirtualConnection` object runs code in a parallel thread,
    which will continue running until it is closed or the Python
    interpreter exits.
    A parallel thread can be closed by calling the :func:`close`
    method of the :class:`VirtualConnection` object that created it.
    This also closes :attr:`user_end` and :attr:`virtual_end`,
    and frees their file descriptors.

    If a :class:`VirtualConnection` object is used in a ``with``
    block in the following manner:

    ::

        with VirtualConnection() as vc:
            # Some code here

    :meth:`close` will be called automatically when the ``with`` block
    is exited.

    If all variables referring to a :class:`VirtualConnection`
    object are removed with ``del`` or by reassigning them,
    the parallel thread will continue to run without a possibility of
    closing it with :meth:`close`.
    In this case, all running instances of the
    :class:`VirtualConnection` class can be closed with

    >>> VirtualConnection.close_all()

    Attributes:
        buffer_size (int):
            The buffer size for the connection
            (how many bits are read at once).

        sleep_time (float):
            How long (in seconds) the object waits after checking for
            input before doing it again.

        process (function):
            The method used for processing input and forming output.
            Its signature should be
            ::

                process(self, input_: bytes) -> output: bytes

            A machine or other device that communicates with its user
            can be simulated by assigning a suitable method to the
            :attr:`process` attribute.

        virtual_end (file-like object):
             This end of the connection is used by :meth:`process` to
             read and write data. It can be written to and read from
             with :func:`os.read` and :func:`os.write`.

        user_end (file-like object):
            This end of the connection is meant to be used by a user to
            send commands to and read data from a simulated device.
            :func:`os.read` doesn't seem to work with it,
            and the `serial <https://pypi.org/project/pyserial/>`_
            module should be used instead.

        thread (:class:`threading.Thread`):
            The parallel thread that runs most functionality in a
            :class:`VirtualConnection` object.

        running_instances:
            Class attribute.
            A set of all currently running instances of the
            :class:`VirtualConnection` class.
    """

    # Keep references to all running instances of this class.
    running_instances = set()

    def __init__(self, process=None, buffer_size=1024, sleep_time=0.01):
        """Initialize a new :class:`VirtualConnection`.

        The new instance starts the parallel thread automatically.

        Args:
            process (function):
                The function assigned to the :attr:`process` attribute.
                If no value is supplied, the :attr:`default_process`
                method will be used instead.

            buffer_size:
                The value of :attr:`buffer_size`.

            sleep_time:
                The value of :attr:`sleep_time`.
        """
        if process:
            self.process = process
        else:
            self.process = self.default_process

        self.buffer_size = buffer_size
        self.sleep_time = sleep_time

        master, slave = pty.openpty()
        # This may be written to and read from with os.write
        # and os.read.
        self.virtual_end = master
        # os.read doesn't seem to work with this, but the serial
        # module does.
        self.user_end = slave

        # A parallel thread for the function self._run().
        self.thread = threading.Thread(target=self._run, args=[])
        # Daemon threads are killed when there are no more non-daemon
        # threads left. Setting daemon to True prevents the parallel
        # thread from running in the background even after the main
        # program has exited.
        self.thread.daemon = True

        # Flag for stopping the parallel thread.
        self._stop_flag = threading.Event()
        # Add this instance to the set.
        self.running_instances.add(self)
        self.thread.start()

    def __enter__(self):
        """Called at the beginning of a ``with`` block; returns
        *self*.
        """
        return self

    def __exit__(self, type_, value, traceback):
        """Called upon exiting a ``with`` block; calls
        :meth:`close`.
        """
        self.close()

    def close(self):
        """Stop the parallel thread and close the connection.

        This function returns only after the parallel thread has
        actually stopped.
        """
        if self.is_running():
            self._stop_flag.set()

        # Wait for the parallel thread to stop.
        while self.is_running():
            time.sleep(self.sleep_time)

        # Unlike set.remove(x), set.discard(x) doesn't raise an error
        # if there is no x in the set.
        self.running_instances.discard(self)

    @classmethod
    def close_all(cls):
        """Close all running instances of this class.

        This function returns only after all parallel threads have
        actually stopped.
        """
        # Iterate over a copy, because closing a VC removes it from
        # cls.running_instances, and removing members from a set
        # during iteration raises an error.
        copy = cls.running_instances.copy()
        for i in copy:
            i.close()

    def is_running(self):
        """Return ``True`` if the parallel thread is running,
        ``False`` otherwise."""
        return self.thread.is_alive()

    @property
    def port(self):
        """Return a device name (e.g. ``'/dev/pts/...'``) that can be
        used as the *port* argument when a :class:`serial.Serial`
        object is created."""
        return os.ttyname(self.user_end)

    def _run(self):
        """Run the parallel thread."""

        while not self._stop_flag.is_set():
            # If there is nothing to read from self.virtual_end,
            # os.read will block forever even with a timeout.
            # select makes sure this doesn't happen.
            ready_for_read, _, _ = select.select([self.virtual_end], [], [], 0)

            if self.virtual_end in ready_for_read:
                input_ = os.read(self.virtual_end, self.buffer_size)
            else:
                time.sleep(self.sleep_time)
                continue

            output = self.process(input_)
            os.write(self.virtual_end, output)

        # Close the files after the parallel thread has stopped.
        # Otherwise the system will run out of file descriptors,
        # if a VirtualConnection is created and stopped many times
        # in a a row.
        os.close(self.user_end)
        os.close(self.virtual_end)

    @staticmethod
    def default_process(input_):
        """Form output based on *input_*.

        This is the default method assigned to :attr:`process`,
        and simply returns *input_* uncanged.

        Args:
            input_: A bytes-like object.

        Returns:
            *input_*.
        """

        output = input_
        return output



"""Unit test for the table module."""
import unittest

from test_turboctl import dummy_parameter
from turboctl import array, table, Types

class Test(unittest.TestCase):
    
    def setUp(self):
        
        # Print long strings when a test fails.
        self.maxDiff = None
        
        p1 = dummy_parameter(
            number=1, 
            name='Parameter 1', 
            indices=range(0), 
            min_=0, 
            max_=65535, 
            default=0, 
            unit='s', 
            writable=True, 
            type_=Types.UINT, 
            size=16, 
            description = 'This is test parameter #1.')
            
        self.row1 = [
            'NUMBER\n1', 
            'NAME\nParameter 1', 
            'INDICES\n-',
            'MIN\n0',
            'MAX\n65535',
            'DEFAULT\n0',
            'UNIT\ns',
            'WRITABLE\nTrue',
            'TYPE\nUnsigned integer',
            'SIZE\n16',
            'DESCRIPTION\nThis is test parameter #1.',
            ]
        self.row1 = [s + '\n\n' for s in self.row1]
            
        p2 = dummy_parameter(
            number=2, 
            name='Parameter 2', 
            indices=range(1,6), 
            min_=-10, 
            max_=10, 
            default=0, 
            unit='Hz', 
            writable=False, 
            type_=Types.SINT, 
            size=32, 
            description = 'This is test parameter #2.')
            
        self.row2 = [
            'NUMBER\n2', 
            'NAME\nParameter 2', 
            'INDICES\n1...5',
            'MIN\n-10',
            'MAX\n10',
            'DEFAULT\n0',
            'UNIT\nHz',
            'WRITABLE\nFalse',
            'TYPE\nSigned integer',
            'SIZE\n32',
            'DESCRIPTION\nThis is test parameter #2.',
            ]
        self.row2 = [s + '\n\n' for s in self.row2]

        self.parameters = {1: p1, 2: p2}
                
    def compare_arrays(self, arr1, arr2):
        for i in range(len(arr2)):
            with self.subTest(i=i):
                for j in range(len(arr2[0])):
                    with self.subTest(i=j):
                        self.assertEqual(arr1[i][j], arr2[i][j])   
        
    def test_all(self):
        arr = array(self.parameters, 'all', {})
        correct_arr = [self.row1, self.row2]
        self.compare_arrays(arr, correct_arr)
        
    def test_numbers(self):
        arr = array(self.parameters, [2,1], {})
        correct_arr = [self.row2, self.row1]
        self.compare_arrays(arr, correct_arr)
        
    def test_print(self):
        string = table(self.parameters, 'all', {})
        correct_string = (
"""
NUMBER  NAME         INDICES  MIN  MAX    DEFAULT  UNIT  WRITABLE  TYPE              SIZE  DESCRIPTION
1       Parameter 1  -        0    65535  0        s     True      Unsigned integer  16    This is test parameter #1.

NUMBER  NAME         INDICES  MIN  MAX    DEFAULT  UNIT  WRITABLE  TYPE              SIZE  DESCRIPTION
2       Parameter 2  1...5    -10  10     0        Hz    False     Signed integer    32    This is test parameter #2.
"""[1:] # Ignore the fiest '\n'.
)
        self.assertEqual(string, correct_string)
        
    def test_line_wrapping(self):
        description = (
            'This is a long string intended to test line wrapping. '
            'This is the secong sentence.\n'
            'This string also contains a manually instered line break. '
            'That is needed to make sure that the line wrapping function '
            'works correctly with existing line breaks.')
        parameter = dummy_parameter(description=description)
        parameters = {1: parameter}
        arr = array(parameters, [1], {'description':35})
        
        string = (
"""DESCRIPTION
This is a long string intended to
test line wrapping. This is the
secong sentence.
This string also contains a
manually instered line break. That
is needed to make sure that the
line wrapping function works
correctly with existing line
breaks.

"""
)     
        self.assertEqual(arr[0][-1], string)
        
        
if __name__ == '__main__':
    unittest.main()


import time
import threading
from dataclasses import dataclass
from typing import Callable, Optional

import serial

from turboctl.telegram import api, codes


@dataclass
class Status:
    """This class stores information about the current status of the pump.

    The values are based on those reported by the pump; if a command is sent to
    the pump to set a value to *X*, the value stored here should only be
    changed once the pump has reported the new changed value.

    Values of ``None`` indicate unknown values that the pump hasn't reported
    yet.
    """

    frequency: Optional[float] = None
    """The stator frequency in Hz."""
    
    temperature: Optional[float] = None
    """The frequency converter temperature in °C."""
    
    current: Optional[float] = None
    """The motor current in A."""
    
    voltage: Optional[float] = None
    """The intermediate circuit voltage in V."""

    pump_on: Optional[bool] = None
    """Indicates whether the pump is on or off."""

    status_bits: Optional[list] = None
    """A list of the status conditions
    (:class:`~turboctl.telegram.codes.StatusBits` members) affecting the pump.
    """

    callback: Optional[Callable] = None
    """A function that is called every time the contents of this object
    are changed. Its signature should be
    ::

        callback(status: Status) -> None
        
    so that it can be called with
    
    ::

        self.callback(self)
        
    Note that since this isn't a bound method, the *self* argument must be
    explicitly passed to the function. 
    """

    def __setattr__(self, name, value):
        """Call :attr:`callback(self) <callback>` whenever an attribute
        is set.
        """
        super().__setattr__(name, value)
        if self.callback:
            self.callback(self)


SERIAL_KWARGS = {
    'port'    : '/dev/ttyACM0',
    'baudrate': 19200,
    'bytesize': serial.EIGHTBITS,
    'parity'  : serial.PARITY_EVEN,
    'stopbits': serial.STOPBITS_ONE,
    'timeout' : 1
}
"""Keyword arguments for creating a :class:`serial.Serial` instance."""


class ControlInterface():
    """This class represents a connection to the pump and can be used to send
    commands to it.
    
    All methods that send commands to the pump return both the query sent to
    the pump and the reply received back as
    :class:`~turboctl.telegram.telegram.TelegramReader` instances.
    
    Attributes:
        status (Status):
            An object that stores the current state of the pump.
            Its attributes are updated every time the pump sends
            back a reply.

        timestep (int or float):
            If *self* was initialized with ``auto_update=True``, this
            determines the time (in seconds) between automatic telegrams.
            The default value is ``1``.

        on_command (bool):
            If this is ``True``, automatic telegrams send the :meth:`pump_on`
            command to the pump instead of :meth:`get_status`.

            See the *auto_update* argument of :meth:`__init__` for further
            details.
            
            The default value is ``None``, which gets updated by the
            :meth:`status` method to match the current state of the pump.
    """

    def __init__(self, port=None, auto_update=False):
        """Initialize a new :class:`ControlInterface`.
        
        Args:
            port (str):
                The device name for the serial port through which the 
                connection is formed. If no value is given, the one defined in
                :const:`SERIAL_KWARGS` is used.

            auto_update (bool):
                If this is ``True``, this object will automatically send a 
                telegram to the pump every :attr:`timestep` seconds and
                update :attr:`status` accordingly. Otherwise :attr:`status`
                will only be updated whenever a command is sent by the user
                calling a method of this class.
                
                If :attr:`status.pump_on <Status.pump_on>` is ``True``,
                the automatic telegram will be :meth:`pump_on` instead of
                :meth:`get_status`.
                This is to prevent the pump from automatically switching off,
                which it does if it hasn't received a :meth:`pump_on` command
                for 10 seconds. 
        """
        self.status = Status()
        self.timestep = 1
        self.on_command = None
        
        kwargs = SERIAL_KWARGS.copy()
        if port:
            kwargs['port'] = port

        # This may raise a serial.SerialException.
        self._connection = serial.Serial(**kwargs)

        # An event flag to close the parallel thread.
        self._stop_flag = threading.Event()

        # _lock prevents messages from being sent at the same time by
        # _run_autoupdate and the user.
        self._lock = threading.Lock()        
        self._thread = threading.Thread(target=self._run_autoupdate,
                                        daemon=True)
        
        # _thread and _lock are created even with auto_update=False,
        # since they are referenced by methods.
        if auto_update:
            self._thread.start()
            
    def __enter__(self):
        """Called upon entering a ``with`` block; returns *self*."""
        return self
    
    def __exit__(self, exc_type, exc_value, traceback):
        """Called upon exiting a ``with`` block; calls :meth:`close`.
        
        The arguments are ignored.
        """
        self.halt()

    def close(self):
        """Close the connection to the pump. If this object was created with
        ``auto_update=False``, the parallel thread sending the update
        telegrams is also closed.
        """
        self._stop_flag.set()
        self._connection.close()

    def _run_autoupdate(self):
        """Periodically send a telegram to the pump and update self.status
        based on the response.
        """
        while not self._stop_flag.is_set():
            if self.on_command:
                self.pump_on()
            else:    
                self.get_status()
            time.sleep(self.timestep)

    def pump_on(self):
        """Turn the pump on.

        This sets :attr:`on_command` to ``True``.
        """
        query, reply = api.status(self._connection, pump_on=True)
        # Only set set self.pump_on=True if the pump actually reports turning
        # on.
        self._update_status(reply)
        self.on_command = True
        return query, reply

    def pump_off(self):
        """Turn the pump off.

        This sets :attr:`on_command` to ``False``.
        """
        query, reply = api.status(self._connection, pump_on=False)
        self._update_status(reply)
        self.on_command = False
        return query, reply

    def get_status(self, pump_on=None):
        """Ask pump status by sending an empty telegram.
        
        If :attr:`on_command` is ``None``, this updates its value to
        :attr:`status.pump_on <Status.pump_on>`.
        This means that a pump that is on when TurboCtl is started stays on,
        and a pump that is off stays off.
        """
        # This is named "get_status" instead of "status", since "status" is
        # already an attribute.
        query, reply = api.status(self._connection)
        self._update_status(reply)

        if self.on_command is None:
            self.on_command = self.status.pump_on

        return query, reply

    def reset_error(self):
        """Reset the error status of the pump."""
        query, reply = api.reset_error(self._connection)
        self._update_status(reply)
        return query, reply

    def read_parameter(self, number, index=0):
        """Read the value of an index of a parameter.
        
        Args:
            number
                The number of the parameter.
                
            index
                The index of the parameter (0 for unindexed parameters).
                
        Raises:
            ValueError:
                If *number* or *index* have invalid values.
        """
        query, reply = api.read_parameter(self._connection, number, index, 
                                          pump_on=self.status.pump_on)
        self._update_status(reply)
        return query, reply
    
    def write_parameter(self, number, value, index=0):
        """Write a value to an index of a parameter.
        
        Args:
            number:
                The number of the parameter.
                                
            value:
                The value to be written.
                
            index:
                The index of the parameter (0 for unindexed parameters).
                
        Raises:
            ValueError:
                If *number*, *value* or *index* have invalid values.
        """
        query, reply = api.write_parameter(self._connection, number, value,
                                           index, pump_on=self.status.pump_on)
        self._update_status(reply)
        return query, reply
        
    def _update_status(self, reply):
        """Update self.status based on the reply from the pump
        (a TelegramReader object).
        """
        self.status.frequency = reply.frequency
        self.status.temperature = reply.temperature
        # The pump reports current in 0.1 A.
        self.status.current = reply.current / 10
        self.status.voltage = reply.voltage
        self.status.status_bits = reply.flag_bits
        if codes.StatusBits.OPERATION in reply.flag_bits:
            self.status.pump_on = True
        else:
            self.status.pump_on = False



"""Run all unit tests at once."""
import unittest
import importlib
import tabulate

# Comment module names away to exclude certain modules 
# from being tested. 
MODS_BY_PKG = {}

MODS_BY_PKG['telegram'] = [
    'test_codes',
    'test_datatypes',
    'test_parser',
    'test_telegram',
    ]

# MODS_BY_PKG['ui'] = [
#     'test_abstractui',
#     'test_command_parser',
#     'test_correct_error_message',
#     'test_output',
#     'test_table',
#     'test_tui',
#     ]

# MODS_BY_PKG['virtualpump'] = [
#     'test_virtualconnection',
#     #'test_virtualpump', 
#     ]

# Uncommenting this module adds 1 successful test, 2 failures, 
# 3 errors and 4 skipped tests. 
# This can be used to test the run_tests() function.

#MODS_BY_PKG['run_tests'] = ['test_run_tests']

class Module():
    """A class representing a single module and its test results."""
    
    def __init__(self, name, package_name):
        self.name = name
        self.package_name = package_name
        
        self.tests_run = 0
        self.failures = []
        self.errors = []
        self.skipped = []
        
    def __add__(self, other):
        """Return *self* + *other*.
        
        If bool(other) is False, return *self*.
        The sum of different test results is acquired by adding 
        all Module objects together.
        """
        
        if not other:
            return self
        
        res = Module('TOTAL', '')
        
        res.tests_run = self.tests_run + other.tests_run
        res.failures = self.failures + other.failures
        res.errors = self.errors + other.errors
        res.skipped = self.skipped + other.skipped
        
        return res
        
    def run_tests(self):
        """Run all tests in this module and record their results."""
        
        module = importlib.import_module(
            f'test_turboctl.{self.package_name}.{self.name}')
        suite = unittest.defaultTestLoader.loadTestsFromModule(module)
        result = unittest.TestResult()    
        suite.run(result)
        
        self.tests_run = result.testsRun
        
        self.failures = [
            FailedResult('failure', *fields) for fields in result.failures]
        
        self.errors = [
            FailedResult('error', *fields) for fields in result.errors]
        
        self.skipped = [
            FailedResult('skipped', *fields) for fields in result.skipped]
        
    @property
    def n_failures(self):
        """The number of failed tests."""
        return len(self.failures)
    
    @property
    def n_errors(self):
        """The number of tests that raised errors."""
        return len(self.errors)
    
    @property
    def n_skipped(self):
        """The number of skipped tests."""
        return len(self.skipped)        

class FailedResult():
    """A class representing an unsuccessful test result of any kind 
    (failed, error or skipped)."""
    
    def __init__(self, type_, name, message):
        """Initialize a new instance.
        
        Args:
            type_: 'failure', 'error' or 'skipped'.
            name: Test name.
            message: Test traceback.
            """
        self.type = type_
        self.name = name
        self.message = message
        
    def __str__(self):
        
        type_str = {'failure': 'Failed ',
                    'error': 'Error in ',
                    'skipped': 'Skipped '}[self.type]        
        
        return f'{type_str}{self.name}\n{self.message}'
    

def run_tests():
    """Run all unit tests and print the results."""
    
    # Sort results by package and module name
    modlist = [Module(mod, pkg) 
               for pkg, mods_in_pkg in sorted(MODS_BY_PKG.items()) 
               for mod in sorted(mods_in_pkg)]
    
    for mod in modlist:
        mod.run_tests()
    
    used_pkgs = set()
    
    def row(module):
        
        if module.package_name in used_pkgs:
            pkg_str = ''
        else:
            pkg_str = module.package_name
            used_pkgs.add(module.package_name)
            
        return [pkg_str, module.name, module.tests_run, 
                module.n_failures, module.n_errors, module.n_skipped]
        
    
    total = None
    for module in modlist:
        total = module + total
    
    modlist.append(total)
    results = [row(module) for module in modlist]
 
    _print_results(results)    
    print()
    _print_messages(modlist)

def _print_results(results):    
    """Print a table displaying the quantities of different test 
    results.
    
    The number of successful tests isn't displayed, since subtests 
    make calculating it difficult. (A successful test with subtests 
    is counted as only one test, but an unsuccesful one contributes 
    one failure per each failed subtest.) 
    """
    
    headers = ['Package', 'Module', 'Total', 'Failures', 'Errors', 'Skipped']
    print(tabulate.tabulate(results, headers=headers))
    
def _print_messages(modlist):
    """Print messages detailing all unsuccessful tests."""
    
    def printall(list_):
        for x in list_:
            print(x)
    
    for mod in modlist:
        printall(mod.failures)
        printall(mod.errors)
        printall(mod.skipped)
    
    
if __name__ == '__main__':
    run_tests()


"""This module contains unit tests for the correct_error_message 
module.
"""
import unittest

from turboctl.ui.correct_error_message import correct_error_message

class TestTooManyArguments(unittest.TestCase):
        
    def test_no_arguments(self):
        string = "cmd_read() takes 1 positional argument but 2 were given"
        new_string = "'read' takes 0 positional arguments but 1 was given"
        self.assertEqual(correct_error_message(string), new_string)
        
    def test_one_argument(self):
        string = "cmd_read() takes 2 positional argument but 3 were given"
        new_string = "'read' takes 1 positional argument but 2 were given"
        self.assertEqual(correct_error_message(string), new_string)
        
    def test_multiple_arguments(self):
        string = "cmd_read() takes 100 positional arguments but 101 were given"
        new_string = ("'read' takes 99 positional arguments "
                      "but 100 were given")
        self.assertEqual(correct_error_message(string), new_string)
        
    def test_variable_number_of_arguments(self):
        string = ("cmd_read() takes from 2 to 3 positional arguments "
                  "but 6 were given")
        new_string = ("'read' takes from 1 to 2 positional "
                      "arguments but 5 were given")
        self.assertEqual(correct_error_message(string), new_string)

class TestOther(unittest.TestCase):

    def test_missing_argument(self):
        
        string = (
            "cmd_read() missing 1 required positional argument: 'number'")
        new_string = (
            """'read' missing 1 required positional argument: 'number'""")
        self.assertEqual(correct_error_message(string), new_string)
        
    def test_unexpected_keyword_argument(self):
        
        string = ("cmd_read() got an unexpected keyword argument 'test'")
        new_string = ("""'read' got an unexpected keyword argument 'test'""")
        self.assertEqual(correct_error_message(string), new_string)
        
    def test_multiple_values_for_same_argument(self):
        
        string = ("cmd_read() got multiple values for argument 'number'")
        new_string = ("""'read' got multiple values for argument 'number'""")
        self.assertEqual(correct_error_message(string), new_string)
    
    
if __name__ == '__main__':
    unittest.main()


"""This module handles parameter access in a
:class:`~turboctl.virtualpump.virtualpump.VirtualPump`.
"""

from turboctl.telegram.codes import (ParameterAccess, ParameterException,
                                     AccessError, CannotChangeError,
                                     MinMaxError, ParameterIndexError,
                                     WrongNumError)
from turboctl.telegram.datatypes import Data
from turboctl.telegram.parser import Parameter


class ParameterComponent():
    """This class defines the part of a
    :class:`~turboctl.virtualpump.virtualpump.VirtualPump` that holds the
    values of pump parameters and handles access to them.
    
    Attributes:
        parameters: A :class:`dict` that represents pump parameters.
            The keys should be parameter numbers (:class:`int`) and values
            corresponding :class:`ExtendedParameter` objects.
    """
    
    def __init__(self, parameters):
        """Initialize a new :class:`ParameterComponent`.
        
        Args:
            parameters: The object to be assigned to :attr:`parameters`.
        """
        self.parameters = parameters
        
    def handle_parameter(self, query, reply):
        """Access a parameter as commanded by *query*.
        
        The data to be returned to the user is written to *reply*.
        
        Args:
            query: A :class:`~turboctl.telegram.telegram.TelegramReader`
                object, which represents the telegram sent to the pump.
            
            reply: A :class:`~turboctl.telegram.telegram.TelegramBuilder`
                object, which is used to construct the telegram sent back
                from the pump.
        """
        if query.parameter_mode != 'none':
            try:
                reply.set_parameter_mode('response')
                reply.set_parameter_value(self._access_parameter(query))
            except ParameterException as error:
                # TODO: Find out how CannotChangeError is used?
                # Should it return 'error' with an error code, or 'no write'.        
                reply.set_parameter_mode('error')
                reply.set_parameter_value(error.member.value)
            
        return reply

    def _access_parameter(self, query):
        """Access a parameter as commanded by *query* and return the return
        value.
        
        Args:
            query: A TelegramReader object.
                
            Raises:
                A subclass of ParameterException: If the parameter cannot be
                    accessed due to a valid readon (i.e. one that can happen
                    to the real pump).
                
                ValueError: If query.parameter_mode is not 'read' or 'write'.
        """
        # Check that the parameter number is valid, and find that parameter.         
        try:
            parameter = self.parameters[query.parameter_number]
        except KeyError:
            raise WrongNumError(
                f'invalid parameter number: {query.parameter_number}')
        # The parameter error messages here are written to descriptive to aid
        # debugging, even though they should never propagate beyond this
        # method. 
            
        # Extract the parameter access code as a string.            
        parameter_code = query.telegram.parameter_code.value
        # Find the corresponding ParameterAccess member.
        access_member = ParameterAccess(parameter_code)
        
        # Check that the access mode matches the size of the parameter.
        bits = access_member.bits
        # The ellipsis means this mode doesn't care about parameter size.
        if bits not in [parameter.bits, ...]:
            raise AccessError(f'bits should be {parameter.bits}, not {bits}')
        
        # Check that the access mode matches the indices of the parameter.
        indexed = access_member.indexed
        # The ellipsis is again a "wild card" value.
        if indexed not in [bool(parameter.indices), ...]:
            raise AccessError('wrong value of *indexed*')
        
        # Some parameters start their indices from some other number than 0.
        # "parameter.indices[0]" is only evaluated here if parameter.indices
        # is not range(0).
        offset = parameter.indices[0] if parameter.indices else 0
        # Example: A parameter has the indices [1, 2, 3] and the values
        # [4, 5, 6]. query.parameter_index is 3.
        # offset is then 1 and index is 3 - 1 = 2, so the value that is
        # returned is parameter.value[2] = 6.
        index = query.parameter_index - offset
                
        # Make sure the index is within range. A try-catch-block could be used
        # instead, but that wouldn't be ideal, since parameter.value is
        # accessed at two different points in the code.
        # This also catches negative indices, but a try-catch-block wouldn't,
        # since negative indices are perfectly valid in Python. 
        if not 0 <= index < len(parameter.value):
            raise ParameterIndexError(
                f'parameter index ({query.parameter_index}) out of range '
                f'({parameter.indices})')
        
        # Write the new value, if the mode is 'write'.
        mode = query.parameter_mode
        if mode == 'write':
            # Check that the parameter is writable.
            if not parameter.writable:
                raise CannotChangeError(
                    f'parameter {query.parameter_number} is not writable')
            
            # Check that the new value is within range.
            
            min_value = parameter.min_value.value
            max_value = parameter.max_value.value
            value = query.parameter_value
            
            if not (min_value <= value <= max_value):
                raise MinMaxError(f'value ({value}) out of range'
                                  f'([{min_value}, {max_value}])')

            # Write the new value as a Data subclass instance of the
            # appropriate type and size.
            parameter.value[index] = (
                parameter.datatype(query.parameter_value, parameter.bits))

        if mode not in ['read', 'write']:
            raise ValueError(f'invalid parameter_mode: {mode}')

        return parameter.value[index].value


class ExtendedParameter():
    """This class represents a parameter that has a value which can 
    change, while the regular :class:`~turboctl.telegram.parser.Parameter`
    class only describes the immutable attributes of a parameter.
    
    Attributes:
        
        number:
            See :attr:`Parameter.number
            <turboctl.telegram.parser.Parameter.number>`.
        
        indices:
            See :attr:`Parameter.indices
            <turboctl.telegram.parser.Parameter.indices>`.
        
        default:
            See :attr:`Parameter.default
            <turboctl.telegram.parser.Parameter.number>`.

        writable:
            See :attr:`Parameter.writable
            <turboctl.telegram.parser.Parameter.number>`.
        
        datatype:
            See :attr:`Parameter.datatype
            <turboctl.telegram.parser.Parameter.number>`.
        
        bits:
            See :attr:`Parameter.bits
            <turboctl.telegram.parser.Parameter.number>`.
            
        value (:class:`turboctl.telegram.datatypes.Data`):
            The current values of the indices of the parameter as a list.
            This will be a list even for unindexed parameters, but in that
            case the length of the list will be 1. The values of the list
            will be instances of
            :attr:`~turboctl.telegram.parser.Parameter.datatype`.
            
        parameters: A :class:`dict` of all instances of this class, 
            with parameter numbers as keys and objects as values.
            This is needed, because the :attr:`min_value` and
            :attr:`max_value` attributes of some parameters depend on the
            values of other parameters.
    """
    
    def __init__(self, parameter, extended_parameters):
        """Initialize a new :class:`ExtendedParameter`.
                
        Args:
            parameter (:class:`~turboctl.telegram.parser.Parameter`):
                The non-extended version of this parameter. Most attributes
                of this object are copied from *parameter*.

            extended_parameters:
                The object to be assigned to :attr:`parameters`. 
        """
        # Copy attributes from *parameter*.
        copied_attributes = [
            'number', 'indices', 'default', 'writable', 'datatype', 'bits'
        ]
        for name in copied_attributes:
            value = getattr(parameter, name)
            setattr(self, name, value)

        self.parameters = extended_parameters
        
        # Save the min and max values of the original parameter.
        # These can be either Data subclass instances or references to the
        # values of other parameters (e.g. 'P18').
        self._raw_min_value = parameter.min_value
        self._raw_max_value = parameter.max_value
        
        # Set self.value based on self.default.
        if type(parameter.default) == list:
            # parameter.default is a list -> indices have different values.
            self.value = parameter.default
        else:
            # parameter.default is a single value -> all indices have the same
            # value.
            if parameter.indices:
                # Indexed parameter.
                self.value = [parameter.default for i in parameter.indices]
            else:
                # Unindexed parameter: parameter.indices = range(0).
                self.value = [parameter.default]
                
    @property
    def min_value(self):
        """Return the minimum value of the parameter.
        
        This is always a :class:`~turboctl.telegram.datatypes.Data` subclass
        instance. If the :attr:`~turboctl.telegram.parser.Paramete.min_value`
        of the original non-extended parameter is a reference to the value of
        another parameter, this property returns that value.
        
        Raises:
            KeyError: If a referenced parameter cannot be found in
                :attr:`parameters`.
        """
        return self._get_true_value(self._raw_min_value)
        
    @property
    def max_value(self):
        """Return the maximum value of the parameter.
        
        This is always a :class:`~turboctl.telegram.datatypes.Data` subclass
        instance. If the :attr:`~turboctl.telegram.parser.Parameter.max_value`
        of the original non-extended parameter is a reference to the value of
        another parameter, this property returns that value.
        
        Raises:
            KeyError: If a referenced parameter cannot be found in
                :attr:`parameters`.
        """
        return self._get_true_value(self._raw_max_value)
        
    def _get_true_value(self, value):
        """Returns the numerical value  of *value*.
        
        If *value* is a reference to the value of another 
        parameter (e.g. 'P18'), this attribute will return the 
        numerical value of that attribute.
        
        The return value will match the format of :attr:`value`.
        
        Args:
            value: An instance of a :class:`turboctl.telegram.datatypes.Data`
                subclass or a string with the format ``'P<number>'``.
        
        Raises:
            ValueError: If *value* is invalid.
        """
        if isinstance(value, Data):
            return value
        
        if isinstance(value, str) and value[0] == 'P':
            # Extract the number.
            num = int(value[1:])
            # Even the values of unindexed parameters are lists.
            # The index is always 0, since there should be no references to
            # unindexed parameters.
            return self.parameters[num].value[0]
        
        raise ValueError(f'invalid *value*: {value}')
                
    def __str__(self):
        """Returns a :class:`str` with the following format:
            
        ::
            
            ExtendedParameter(
                number=<number>,   
                indices=<indices>,
                min_value=<min_value>,
                max_value=<max_value>,
                default=<default>,
                writable=<writable>,
                datatype=<datatype>,
                bits=<bits>,
                value=<value>
            )
            
        If the :attr:`~turboctl.telegram.parser.Parameter.min_value` or 
        :attr:`~turboctl.telegram.parser.Parameter.max_value` attributes of the
        original parameter are references, both the reference and the actual
        value are displayed. The following is an example of the format:
        
        ::
            
            ...
            min_value='P18' (Uint(1, 16)),
            ...
        """
        
        if isinstance(self._raw_min_value, str):
            min_str = repr(self._raw_min_value) + f' ({self.min_value})'
        else:
            min_str = str(self.min_value)
            
        if isinstance(self._raw_max_value, str):
            max_str = repr(self._raw_max_value) + f' ({self.max_value})'
        else:
            max_str = str(self.max_value)
        
        return f"""
ExtendedParameter(
    number={self.number},   
    indices={self.indices},
    min_value={min_str},
    max_value={max_str},
    default={self.default},
    writable={self.writable},
    datatype={self.datatype.__name__},
    bits={self.bits},
    value={self.value}
)
"""[1:-1]


class ExtendedParameters(dict):
    """A :class:`dict` of :class:`ExtendedParameter` objects.
    
    This class can be used to avoid the need to manually initialize a
    :class:`dict` of :class:`ExtendedParameter` objects. After initialization
    instances of this class behave like regular :class:`dict` objects.  
    """

    def __init__(self, parameters):
        """Initialize a new :class:`ExtendedParameters` object from
        *parameters* (a :class:`dict` of
        :class:`~turboctl.telegram.parser.Parameter` objects).
        
        The data from each :class:`~turboctl.telegram.parser.Parameter` object
        is copied into an :class:`ExtendedParameter` object. The objects are
        initialized in such an order that no errors will be raised because of
        references to uninitialized parameters.
        """
        super().__init__()        

        max_iters = 5
        # The min and max values of some parameters depend on the 
        # values of other parameters, and so some parameters cannot be 
        # initialized on the first iteration, or the first several 
        # iterations, in case their dependencies also have 
        # dependencies.
        # In practice two iterations seems to be enough if real pump 
        # parameters are used, but there could in theory be long 
        # dependency chains that need more iterations.
        for i in range(max_iters):
            for num, p in parameters.items():
                if num not in self.keys():
                    try:
                        self[num] = ExtendedParameter(p, self)
                    # A KeyError is raised if a parameter's dependency 
                    # has not been initialized.
                    except KeyError:
                        pass



import unittest
import time
import serial

from turboctl import (AbstractUI, VirtualPump, VirtualConnection, Query, 
                      TelegramWrapper, Types, PARAMETERS, ParameterError, 
                      StatusBits)
from test_turboctl import dummy_parameter

# Add some new parameters for simpler testing.
# The largest actual parameter number is 1102, so these won't 
# override any old parameters. Otherwise the virtual pump might not 
# work, since it accesses hardware parameters.
new_parameter_list = [
    dummy_parameter(number=2001, type_=Types.UINT),
    dummy_parameter(number=2002, type_=Types.SINT, min_ = -1000),
    dummy_parameter(number=2003, type_=Types.FLOAT, min_ = -1000),
    dummy_parameter(number=2004, indices=range(5)),
    dummy_parameter(number=2005, writable=False),
    dummy_parameter(number=2006, max_=10),
]


class Base(unittest.TestCase):
    
    @classmethod
    def setUpClass(cls):
        # Replace *PARAMETERS* with *new_parameters*
        new_parameters = PARAMETERS 
        new_parameters.update({p.number: p for p in new_parameter_list})
        TelegramWrapper.parameters = new_parameters
        
        cls.pump = VirtualPump(new_parameters)
        cls.ui = AbstractUI(cls.pump.port)
        
    @classmethod
    def tearDownClass(cls):
        cls.pump.close()
        # Reset TelegramWrapper back to its original state to avoid 
        # breaking tests in other test modules when running all unit 
        # tests at once.
        TelegramWrapper.parameters = PARAMETERS
                
        
class TestReadAndWriteParameter(Base):
    
    def read_and_write(self, value, number, index):
        q, r = self.ui.write_parameter(value, number, index)
        
        self.assertEqual(q.parameter_mode, 'write')
        self.assertEqual(q.parameter_number, number)
        self.assertEqual(q.parameter_index, index)
        self.assertAlmostEqual(q.parameter_value, value, 5)
        
        self.assertEqual(r.parameter_mode, 'response')
        self.assertEqual(r.parameter_number, number)
        self.assertEqual(r.parameter_index, index)
        self.assertAlmostEqual(r.parameter_value, value, 5)
        
        q, r = self.ui.read_parameter(number, index)
        
        self.assertEqual(q.parameter_mode, 'read')
        self.assertEqual(q.parameter_number, number)
        self.assertEqual(q.parameter_index, index)
        self.assertAlmostEqual(q.parameter_value, 0, 5)
        
        self.assertEqual(r.parameter_mode, 'response')
        self.assertEqual(r.parameter_number, number)
        self.assertEqual(r.parameter_index, index)
        self.assertAlmostEqual(r.parameter_value, value, 5)
        
    def test_uint(self):
        self.read_and_write(100, 2001, 0)
        
    def test_sint(self):
        self.read_and_write(-100, 2002, 0)
        
    def test_float(self):
        # This also tests reading 32-bit numbers.
        self.read_and_write(123.456, 2003, 0)
        
    def test_indexed(self):
        self.read_and_write(123, 2004, 3)
        
    def test_no_write(self):
        q, r = self.ui.write_parameter(123, 2005, 0)
        
        self.assertEqual(r.parameter_mode, 'no write')
        self.assertEqual(r.parameter_number, 2005)
        self.assertEqual(r.parameter_index, 0)
        self.assertEqual(r.parameter_value, 123)
                
    def test_error(self):
        q, r = self.ui.write_parameter(11, 2006, 0)
        self.assertEqual(r.parameter_mode, 'error')
        self.assertEqual(r.error_code, ParameterError.MINMAX)

        
class TestOtherCommands(Base):
    
    def test_on_off_and_status(self):
        
        # Make sure the status signal is correct, and the pump is 
        # initially not turning.
        q, r = self.ui.status()
        status_bits = [2,22,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,20]
        status_query = Query(status_bits)
        self.assertEqual(q, status_query)        
        self.assertFalse(StatusBits.TURNING in r.control_or_status_set)

        # Make sure the on_off signal is correct.
        q, r = self.ui.on_off()
        onoff_bits = [2,22,0,0,0,0,0,0,0,0,0,4,1,0,0,0,0,0,0,0,0,0,0,17]
        onoff_query = Query(onoff_bits)
        self.assertEqual(q, onoff_query)
        
        # Wait for the pump to start.
        time.sleep(0.1)
        
        # Make sure the pump starts turning.
        q, r = self.ui.status()
        self.assertTrue(StatusBits.TURNING in r.control_or_status_set)
        
    def test_save_data(self):
        # Turn the pump on.
        self.ui.on_off()
        
        # Write a value and restart the pump.
        self.ui.write_parameter(200, 1)
        self.ui.on_off()
        self.ui.on_off()
        # The written value is replaced by the default value.
        q, r = self.ui.read_parameter(1)
        self.assertEqual(r.parameter_value, 180)
        
        # The written value is saved to nonvolatile memory.
        self.ui.write_parameter(200, 1)
        self.ui.save_data()
        self.ui.on_off()
        self.ui.on_off()
        q, r = self.ui.read_parameter(1)
        self.assertEqual(r.parameter_value, 200)
        
    def test_set_frequency(self):
        # Turn the pump on.
        self.ui.on_off()
        
        # Make sure the setpoint frequency is the default value.
        q, r = self.ui.read_parameter(24)
        self.assertEqual(r.parameter_value, 1000)
        
        # Set a new setpoint.
        self.ui.set_frequency(800)
        time.sleep(0.1)
        
        # Make sure the change was applied.
        q, r = self.ui.read_parameter(24)
        self.assertEqual(r.parameter_value, 800)
        
        
class TestConnection(unittest.TestCase):
    
    @classmethod
    def setUpClass(cls):
        cls.vc = VirtualConnection()
        cls.ui = AbstractUI(cls.vc.port)
        
    @classmethod
    def tearDownClass(cls):
        cls.vc.close()
    
    def test_invalid_port(self):
        with self.assertRaises(serial.SerialException):
            AbstractUI('test')
            
    def test_empty_port(self):
        ui = AbstractUI(None)
        with self.assertRaises(serial.SerialException):
            ui.status()
            
    def test_no_response(self):
        def process(input_):        
            return b''
        self.vc.process = process
        with self.assertRaises(ValueError):
            self.ui.status()
            
    def test_invalid_response(self):
        def process(input_):        
            return b'1234'
        self.vc.process = process
        with self.assertRaises(ValueError):
            self.ui.status()

        
if __name__ == '__main__':
    unittest.main()


"""This module uses the `urwid <http://urwid.org/>`_ library to create
widgets (see :class:`urwid.Widget`) for a text-based user interface
containing a scrollable command line interface below a text screen.

The docstring of each widget defined here tells whether it is a box,
flow or fixed widget. The differences between these widget types are
explained `here <http://urwid.org/manual/widgets.html
#box-flow-and-fixed-widgets>`_.
"""

# pylint: disable=too-many-arguments
# Some urwid methods have many arguments.

# pylint: disable=unused-argument
# Subclasses can't safely change the signatures of methods inherited
# from a superclass.

# pylint: disable=protected-access
# urwid documentation suggests using the _invalidate method.

import urwid


class ScrollableCommandLines(urwid.WidgetWrap):
    """A widget consisting of a scrollable command line interface and
    a scroll bar with two arrow buttons for scrolling it.

    This is a box widget; its container sets it height and width.

    Attributes:
        position (:class:`Position`):
            Keeps track of how far the screen is scrolled.

        command_lines (:class:`CommandLines`):
            Provides the command line interface.

        scroller (:class:`Scroller`):
            Adds scrolling functionality to :attr:`command_lines`.

        scrollbar (:class:`ScrollBar`):
            The scroll bar.

        arrow_up (:class:`ScrollButton`):
            An arrow button that can be clicked to scroll the screen
            up.

        arrow_down (:class:`ScrollButton`):
            An arrow button that can be clicked to scroll the screen
            down.
    """

    def sizing(self):  # pylint: disable=empty-docstring
        """"""
        # By default, self.sizing() seems to return the sizing of
        # urwid.Columns, frozenset({'box', 'flow'}).
        # Overriding _sizing doesn't work, so this method has to be
        # defined.
        # The """""" docstring prevents this method from inheriting the
        # docstring of urwid.Widget.sizing.
        # Because the docstring is empty, this method isn't included in
        # the documentation.
        return {urwid.BOX}

    def __init__(self, inputfile, outputfile):
        """Initialize a new :class:`ScrollableCommandLines`.

        The arguments are passed to the initializer of
        :attr:`command_lines`.
        """
        self.position = Position()

        self.command_lines = CommandLines(inputfile, outputfile)
        self.scroller = Scroller(self.command_lines, self.position)
        self.scrollbar = ScrollBar(self.position)
        self.arrow_up = ScrollButton('▲', -1, self.position)
        self.arrow_down = ScrollButton('▼', +1, self.position)

        self.position.listeners = {self.command_lines, self.scrollbar}

        # (num, widget)-tuples are used to specify the height
        # (for Piles) or width (for Columns) of contained widgets.
        pile = urwid.Pile([(1, self.arrow_up), self.scrollbar,
                           (1, self.arrow_down)])
        super().__init__(urwid.Columns([self.scroller, (2, pile)]))


class CommandLines(urwid.WidgetWrap):
    """This widget provides a terminal-like command line interface.

    Entering commands scrolls the screen automatically so that the
    latest command is always visible on the screen, but the
    functionality to scroll up to view old commands is provided by the
    :class:`Scroller` class instead of this class.

    This is a flow widget; its container sets it height but not width.

    Attributes:
        inputfile (file-like object):
            Input entered into the command-line interface by the user
            is written to this object.

        outputfile (file-like object):
            The command-line interface reads its output from this
            object and prints it on the screen.

        history (:class:`CommandHistory`):
            This object stores previously given commands.

        edit (:class:`urwid.Edit`):
            This widget provides the functionality for displaying and
            editing text.
    """

    def __init__(self, inputfile, outputfile):
        """Initialize a new :class:`CommandLines` object.

        The arguments set the values of :attr:`inputfile` and
        :attr:`outputfile`.
        """
        self.inputfile = inputfile
        self.outputfile = outputfile
        self.history = CommandHistory()
        self.edit = urwid.Edit()
        super().__init__(self.edit)

    def move_cursor_to_end(self):
        """Move the cursor to the end of the editable string."""
        self.edit.set_edit_pos(len(self.edit.edit_text))

    def update(self):
        """Print the string returned by
        :attr:`outputfile.read() <outputfile>` to the screen.
        """
        string = self.outputfile.read()
        if string:
            self.edit.set_caption(self.edit.caption + string)

    def history_up(self):
        """Scroll command history up one step.

        Calling this function has approximately the same effect as
        pressing the up arrow button in a Linux terminal.
        """
        self.history.update_command(self.edit.edit_text)
        self.history.up()
        self.edit.set_edit_text(self.history.get_command())
        self.move_cursor_to_end()

    def history_down(self):
        """Scroll command history down one step.

        Calling this function has approximately the same effect as
        pressing the down arrow button in a Linux terminal.
        """
        self.history.update_command(self.edit.edit_text)
        self.history.down()
        self.edit.set_edit_text(self.history.get_command())
        self.move_cursor_to_end()

    def enter(self):
        """Enter the current editable string as a command."""
        command = self.edit.edit_text
        self.history.update_command(command)
        self.history.enter_command()
        self.inputfile.write(command + '\n')
        self.edit.set_caption(self.edit.caption + command + '\n')
        self.edit.set_edit_text('')

    def keypress(self, size, key):
        """Handle key presses.

        The up and down arrow keys are used to browse command history.
        Enter enters a command.

        Returns:
            ``None`` if the key press was handled by this widget,
            *key* if it was not.
        """
        if key == 'enter':
            self.enter()
            return None

        if key == 'up':
            self.history_up()
            return None

        if key == 'down':
            self.history_down()
            return None

        return super().keypress(size, key)


class CommandHistory:
    """A class to store command history.

    This class emulates the command history browsing behaviour of a
    typical Linux terminal.

    Attributes:
        history:
            A list of previously entered commands.
            Index 0 is set to ``None`` to give :attr:`history`
            and :attr:`temp_history` the same length.

        temp_history:
            The same as :attr:`history`, but includes the newest
            command (the one that is currently being written) at
            index 0.
            If any previously given commands are modified
            (through :meth:`update_command`), the modified versions
            are saved here, while :attr:`history` contains the
            original ones.

        index:
            The index of the currently selected command.
            ``0`` is the command currently being written;
            ``1`` is the previous one etc.
    """

    def __init__(self):
        """Initialize :attr:`history` to ``[]``,
        :attr:`temp_history` to ``[None]`` and
        :attr:`index` to ``0``.
        """
        self.history = [None]
        self.temp_history = ['']
        self.index = 0

    def up(self):  # pylint: disable=invalid-name
        """Increase :attr:`index` by 1 (i.e. select a command that is
        older than the current one).

        If the current command is the oldest one, nothing is done.
        """
        self.index += 1
        if self.index >= len(self.history):
            self.index = len(self.history) - 1

    def down(self):
        """Decrease :attr:`index` by 1 (i.e. select a command that is
        newer than the current one).

        If the current command is the newest one, nothing is done.
        """
        self.index -= 1
        if self.index < 0:
            self.index = 0

    def get_command(self):
        """Return the currently selected command as a string."""
        return self.temp_history[self.index]

    def update_command(self, string):
        """Change the currently selected command into *string*."""
        self.temp_history[self.index] = string

    def enter_command(self):
        """Enter the currently selected command.

        This saves the command into memory and  and resets
        :attr:`index` to 0. If the command was created by modifying a
        previous command, that command's history entry will be reset
        into its original form. If multiple identical commands are
        entered consecutively, only one is saved; empty commands
        are not saved at all.
        """
        # The command that was entered:
        entered_cmd = self.temp_history[self.index]
        # The command before that:
        try:
            previous_cmd = self.temp_history[self.index + 1]
        except IndexError:
            previous_cmd = None

        # If a previous command was modified and the entered, changes
        # to it in the command history are erased by copying the
        # original form from *history*.
        if self.index > 0:
            self.temp_history[self.index] = self.history[self.index]

        # Save the entered command to the command history.
        # However, don't save consecutive duplicates or empty commands.
        if entered_cmd not in (previous_cmd, ''):
            self.history.insert(1, entered_cmd)
            self.temp_history.insert(1, entered_cmd)

        # Add a new command to be edited; initially this is empty.
        self.history[0] = ''

        # Reset *index*.
        self.index = 0


class Scroller(urwid.WidgetWrap):
    """A container widget that makes a flow widget scrollable.

    This is a box widget; its container sets it height and width.

    Attributes:
        widget (urwid.Widget):
            The widget to be made scrollable. This should be a flow
            widget.

        position (:class:`Position`):
            An object to keep track of how far the widget is scrolled.
            The same :class:`Position` object should be shared among all
            objects that can scroll the widget (e.g. the scroll bar).
    """

    def __init__(self, widget, position):
        """Initialize a new :class:`Scroller`.

        The arguments set the values of :attr:`widget` and
        :attr:`position`.
        """
        self.widget = widget
        self.position = position
        super().__init__(urwid.Filler(self.widget))

    def render(self, size, focus=False):
        """Render the visible portion of the scrollable widget,
        update :attr:`position` and
        return an :class:`urwid.Canvas` object.
        """
        cols = size[0]
        # self.widget should be a float widget.
        canvas = self.widget.render((cols,), focus)
        # *canvas* has been finalized and can't be edited.
        new_canvas = urwid.CompositeCanvas(canvas)

        # How many rows should be rendered.
        given_rows = size[1]
        # How many rows tall self.widget is.
        canvas_rows = new_canvas.rows()

        # Update self.position
        self.position.visible_rows = given_rows
        self.position.total_rows = max(canvas_rows, given_rows)

        # Compute how many rows should be trimmed from the canvas
        # (or added to it as padding).
        total_trim = canvas_rows - given_rows
        top_trim = round(self.position.relative * total_trim)
        # A negative trim means padding, but padding should be added
        # to the bottom, so *top_trim* can never be negative
        if top_trim < 0:
            top_trim = 0

        # If *bottom_trim* is negative, the bottom is padded instead
        # of trimmed.
        bottom_trim = total_trim - top_trim

        # If the cursor is moved off the screen, it shouldn't be
        # displayed.
        if new_canvas.cursor:
            # canvas.cursor is None or (col, row).
            cursor_row = new_canvas.cursor[1]
            if cursor_row >= given_rows + top_trim:
                new_canvas.cursor = None

        # Negative arguments trim, positive arguments pad
        # -> trim amounts must be converted to negative values.
        new_canvas.pad_trim_top_bottom(-top_trim, -bottom_trim)
        return new_canvas

    def mouse_event(self, size, event, button, col, row, focus):
        """Handle mouse events.

        The mouse wheel scrolls the terminal up and down.

        Returns:
            ``True`` if the event was handled by this widget,
            ``False`` otherwise.
        """
        # Mouse wheel up
        if button == 4:
            self.position.absolute -= 1
            return True

        # Mouse wheel down
        if button == 5:
            self.position.absolute += 1
            return True

        return super().mouse_event(size, event, button, col, row, focus)

    def keypress(self, size, key):
        """Handle key presses.

        Pressing any key scrolls the terminal to the bottom so that
        the current command can be edited.

        Returns:
            ``None`` if the key press was handled by this widget,
            *key* if it was not.
        """
        self.position.relative = 1
        return super().keypress(size, key)


class Position:
    """A class to keep track of the position to which a widget made
    scrollable with :class:`Scroller` has been scrolled.

    Attributes:
        listeners (iterable):
            The widgets that should be re-rendered whenever
            :attr:`absolute` or :attr:`relative` is changed.

        total_rows (int):
            The height of the scrollable widget in rows,
            including those that are not currently visible on the
            screen.
            Empty rows are also included.
            The :class:`Scroller` widget should update this attribute
            whenever it is rendered.

        visible_rows (int):
            Like :attr:`total_rows`, but only counts the rows visible
            on the screen.

        relative (float):
            A relative scroll position between ``0`` (scrolled to the
            top) and ``1`` (scrolled to the bottom).
            Changing this also changes the value of :attr:`absolute`.
            Values below ``0`` are set to ``0`` and values above ``1``
            to ``1``.

        absolute (int):
            An absolute scroll position given as the number of
            invisible rows that are located above the top edge of the
            widget.
            Changing this also changes the value of :attr:`relative`.
            Values below ``0`` are set to ``0`` and values above
            :attr:`max_absolute` to :attr:`max_absolute`.
    """

    def __init__(self, listeners=None):
        """Initialize a new :class:`Position` object.

        Args:
            listeners (iterable):
                The initial value for :attr:`listeners`.
                This may be changed later. If no value is given,
                :attr:`listeners` is initialized to ``set()``.
        """
        if listeners is not None:
            self.listeners = listeners
        else:
            self.listeners = set()
        self.total_rows = 1
        self.visible_rows = 1
        self._relative = 0

    @property
    def max_absolute(self):
        """The maximum value of :attr:`absolute`, equal to
        ``total_rows - visible_rows``."""
        ret = self.total_rows - self.visible_rows
        if ret < 0:
            raise AssertionError(
                f'*total_rows* should always be >= *visible_rows*; '
                f'now they were {self.total_rows} and {self.visible_rows}.')
        return ret

    @property
    def relative(self):
        return self._relative

    @relative.setter
    def relative(self, value):
        if value < 0:
            value = 0

        if value > 1:
            value = 1

        self._relative = value
        for listener in self.listeners:
            listener._invalidate()

    @property
    def absolute(self):
        return round(self._relative * self.max_absolute)

    @absolute.setter
    def absolute(self, value):
        if value < 0:
            value = 0

        if value > self.max_absolute:
            value = self.max_absolute

        if self.max_absolute != 0:
            self._relative = value / self.max_absolute
        else:
            self._relative = 0

        for listener in self.listeners:
            listener._invalidate()


class ScrollBar(urwid.WidgetWrap):
    """A widget that provides a vertical scroll bar.

    This is a box widget; its container sets it height and width.

    Attributes:
        position (:class:`Position`):
            An object used to keep track of how far the screen is
            scrolled.

        scroller_char (single-character :class:`str`):
            The character used to draw the moving part of the scroll
            bar.

        background_char (single-character :class:`str`):
            The character used to draw the other parts of the scroll
            bar.
    """

    def __init__(self, position, scroller_char='█', background_char='░'):
        """Initialize a new scrollbar.

        The arguments set the values of :attr:`position`,
        :attr:`scroller_char` and :attr:`background_char`.
        """
        self.position = position
        self.scroller_char = scroller_char
        self.background_char = background_char
        # A filler must be used to make this a box widget.
        widget = urwid.Filler(urwid.Text(''))
        super().__init__(widget)

    def render(self, size, focus=False):
        """Generate the scroll bar and render it."""
        text = self._generate_text(size)
        self._w.original_widget.set_text(text)
        return super().render(size, focus)

    def _generate_text(self, size):
        """Generate the text displayed by *self.*

        Returns:
            A list of strings, each corresponding to a text row.
            The list has the dimensions indicated by *size*.
        """
        scroller_relative_size = (
            self.position.visible_rows / self.position.total_rows)

        if scroller_relative_size > 1:
            scroller_relative_size = 1

        ncols, nrows = size
        n_scroller_rows = round(nrows * scroller_relative_size)

        if n_scroller_rows < 1:
            n_scroller_rows = 1

        n_background_rows = nrows - n_scroller_rows
        n_background_rows_above = (
            round(self.position.relative * n_background_rows))
        n_background_rows_below = n_background_rows - n_background_rows_above

        basic_row = ncols * self.background_char
        scroller_row = ncols * self.scroller_char

        return (n_background_rows_above * [basic_row]
                + n_scroller_rows * [scroller_row]
                + n_background_rows_below * [basic_row])

    def mouse_event(self, size, event, button, col, row, focus):
        """Handle mouse events.

        Mouse button 1 moves the scroller so that clicking the scroll
        bar at a point *X* % from its top will set
        :attr:`position`:attr:`.relative <Position.relative>`
        to ``X / 100``.

        Returns:
            ``True`` if the event was handled by this widget,
            ``False`` otherwise.
        """
        if button == 1:
            nrows = size[1]  # size = (ncols, nrows)
            max_row = nrows - 1
            self.position.relative = row / (max_row)

        return True


class ScrollButton(urwid.WidgetWrap):
    """A widget that can be clicked with the mouse to scroll another
    widget.

    This is a box widget; its container sets it height and width.

    Attributes:
        step (int):
            The amount of rows scrolled when this widget is clicked.
            Negative values are used to scroll towards the top of the
            scrollable widget, positive values towards the bottom.

        position (:class:`Position`):
            An object used to keep track of how far the scrollable
            widget is scrolled.
    """

    def __init__(self, symbol, step, position):
        """Initialize a new :class:`ScrollButton`.

        Args:
            symbol (str):
                The character or string that is displayed on the
                button.

            step:
                The value of :attr:`step`.

            position:
                The value of :attr:`position`.
        """
        self.position = position
        self.step = step
        widget = urwid.Filler(urwid.Text(symbol))
        super().__init__(widget)

    def mouse_event(self, size, event, button, col, row, focus):
        """Handle mouse events.

        A left click scrolls the scrollable widget by :attr:`step`
        rows.

        Returns:
            ``True`` if the event was handled by this widget,
            ``False`` otherwise.
        """
        if button == 1:
            self.position.absolute += self.step

        return True


"""This is the main script used to run TurboCtl."""

import argparse
import sys

from turboctl.ui import status_format
from turboctl.ui.command_line_ui import CommandLineUI
from turboctl.ui.docs import docs
from turboctl.ui.queuefile import QueueFile
from turboctl.virtualpump.virtualpump import VirtualPump
from test_turboctl.run_tests.run_tests import run_tests


### Command-line arguments ###

# The argparse module hepls with parsing command line arguments.
# The -h option displays an automatically generated help message.
parser = argparse.ArgumentParser(description='This script runs TurboCtrl')

# TurbCtl can be run in four modes: show docs (-d), run tests (-t),
# run without urwid (-s) or run with the regular UI (no letter).
mode_group = parser.add_mutually_exclusive_group()
mode_group.add_argument('-d', '--docs',
                        help='open documentation in a web browser and exit',
                        action='store_true')
mode_group.add_argument('-t', '--test', help='run tests', action='store_true')
mode_group.add_argument('-s', '--simple',
                        help="use a UI that doesn't require urwid", 
                        action='store_true')

# Regardless of mode, TurboCtl can be run with a virtual pump (-v) 
# or with a real pump (-p and port name). These options don't matter when
# running in doc/test mode.
vpump_group = parser.add_mutually_exclusive_group()
vpump_group.add_argument('-v', '--virtual', help='use a virtual pump',
                         action='store_true')
vpump_group.add_argument('-p', '--port', 
                         help='the address of the serial port device')

# Automatic polling can be disabled with this argument in all modes (even
# though this doesn't do anything in doc/test mode).
parser.add_argument('-n', '--no-poll',
                    help=('do not poll the status of the pump by sending '
                          'automatic telegrams'), 
                    action='store_true')

args = parser.parse_args()


### The main part of the script ###

def main():
    """Execute the script."""

    if args.docs:
        docs()
        return

    if args.test:
        run_tests()
        return
    
    try:
        # If virtualpump was defined in another function
        # and an error occurred between its creation and the
        # 'return virtualpump' statement,
        # the finally block wouldn't work, because 'virtualpump' wouldn't
        # be included in the namespace of main().
        # This also holds for command_line_ui. Because of this,
        # both objects are defined directly in main().
        virtualpump = VirtualPump() if args.virtual else None
        port = get_port(virtualpump)

        # Automatically poll the pump only if the advanced UI is used,
        # since the simple UI doesn't contain a screen to display the changing
        # status.
        autoupdate = not args.no_poll
        inputfile, outputfile = get_io_files()
        command_line_ui = CommandLineUI(port, autoupdate,
                                        inputfile, outputfile)
        ui = get_ui(command_line_ui)
        ui.run()
    finally:
        # Make sure all parallel threads are closed, even if the program
        # crashes. Otherwise the terminal may be left in a weird unresponsive
        # state.

        try:
            # This raises a NameError if the script raises an error
            # before command_line_ui is defined.
            command_line_ui.control_interface.close()
        except NameError:
            pass

        try:
            # An AttributeError is raised if virtualhv is None.
            virtualpump.connection.close()
        except AttributeError:
            pass


def get_port(virtualpump):
    """Return the port used for the serial connection.

    If a virtual pump isn't used, the *virtualpump* argument is ignored.
    """
    if args.virtual:
        return virtualpump.connection.port
    if args.port:
        return args.port
    # Use the default value.
    return None


def get_io_files():
    """Return the input and output files for the UI.

    Returns:
        (inputfile, outpufile)
    """
    if args.simple:
        inputfile = sys.stdin
        outputfile = sys.stdout
    else:
        inputfile = QueueFile(block=True)
        outputfile = QueueFile()

    return inputfile, outputfile


def get_ui(command_line_ui):
    """Return *command_line_ui* or an AdvancedUI object that uses
    *command_line_ui*.
    """
    if args.simple:
        ui = command_line_ui
    else:
        # Because AdvancedTUI imports urwid, it is imported only if
        # it is needed.
        # This makes it possible to run HVCtl without urwid by using
        # CommandLineUI.

        # Monkey-patch a bug in Urwid before using it anywhere else.
        from turboctl.hacks import monkeypatch_urwid
        monkeypatch_urwid()

        from turboctl.ui.advanced_tui import AdvancedTUI

        advanced_ui = AdvancedTUI(command_line_ui.run,
                                  command_line_ui.inputfile,
                                  command_line_ui.outputfile,
                                  status_format.palette)

        # Show the status of the HV generator in the UI.
        def callback(status):
            text = status_format.status_screen(status)
            advanced_ui.display.set_text(text)
        command_line_ui.control_interface.status.callback = callback
        ui = advanced_ui

    return ui


if __name__ == '__main__':
    main()



import unittest
import serial
import time
import threading

from turboctl.telegram.codes import ControlBits
from turboctl.telegram.telegram import TelegramBuilder, TelegramReader
from turboctl.virtualpump.virtualpump import VirtualPump


class TestStop(unittest.TestCase):
    """Make sure VirtualPump shuts down all parallel threads properly."""
    
    def test_with_block(self):
        """Test the __enter__ and __exit__ methods."""
        
        # Wait for parallel threads to shut down in case some other tests are
        # run before this one.
        # 0.2 s is twice the default timestep of the frequency loop
        # in HardwareComponent.
        time.sleep(0.2)
        # Count the number of active threads.
        original_count = threading.active_count()
        
        with VirtualPump():
            # Creating a VirtualPump should increase the number of threads by
            # two (one from the connection thread and one from the frequency
            # simulation thread).
            self.assertEqual(threading.active_count(), original_count + 2)
            
        # Exiting the with block should automatically stop the new threads.
        # Wait for them to shut down and make sure they have done so.
        time.sleep(0.2)
        self.assertEqual(threading.active_count(), original_count)
        
    def test_stop(self):
        """Test manually stopping VirtualPump."""
        
        # Do the same things as in test_with_block, but stop the VirtualPump
        # manually.
        time.sleep(0.2)
        original_count = threading.active_count()
        
        vp = VirtualPump()
        self.assertEqual(threading.active_count(), original_count + 2)
            
        vp.stop()
        time.sleep(0.2)
        self.assertEqual(threading.active_count(), original_count)


class TestVirtualPump(unittest.TestCase):
    
    def setUp(self):
        self.vp = VirtualPump()
        # A timeout is needed; otherwise the connection will block forever if
        # the pump doesn't send a response.
        self.connection = serial.Serial(self.vp.connection.port, timeout=0.01)
        
    def tearDown(self):
        self.vp.stop()
        self.connection.close()
        
    def send(self, mode='none', number=0, value=0, bits=[]):
        
        query = (TelegramBuilder()
                 .set_parameter_mode(mode)
                 .set_parameter_number(number)
                 .set_parameter_value(value)
                 .set_flag_bits(bits)
                 .build('query'))
        
        self.connection.write(bytes(query))
        time.sleep(0.1)
        bytes_out = self.connection.read(24)
        reply = TelegramBuilder().from_bytes(bytes_out).build('reply')
        return TelegramReader(reply, 'reply')
        
    def test_parallel_thread_starts_and_stops(self):
        with VirtualPump() as vp:
            self.assertTrue(vp.connection.is_running())
        
        time.sleep(2 * vp.connection.sleep_time)
        # The parallel thread must finish the current iteration before
        # stopping.
        self.assertFalse(vp.connection.is_running())
    
    def test_invalid_telegram(self):
        """Make sure sending something that isn't a valid telegram 
        doesn't cause blocking or other unwanted effects.
        """
        self.connection.write(bytes([1,2,3,4]))
        out = self.connection.read(24)
        self.assertEqual(out, bytes())

    def test_change_setpoint(self):
        """Make sure ParameterComponent and HardwareComponent work properly
        together. The frequency setpoint parameter should determine the
        pump frequency, which in turn should be reflected in the frequency
        parameter.
        """
        # Make the pump accelerate infinitely fast in order to make testing
        # easier.
        self.vp.hardware_component.abs_acceleration = float('inf')
        # The bits required to turn or keep the pump on.
        bits = [ControlBits.COMMAND, ControlBits.ON]
        
        # Ask pump frequency and turn it on.
        reply = self.send(mode='read', number=3, bits=bits)
        self.assertEqual(reply.parameter_value, 0)
        self.assertEqual(reply.frequency, 0)
        
        # Wait for the pump to turn on.
        time.sleep(2 * self.vp.hardware_component.step)
        
        # Ask pump frequency, which should now be at the setpoint.
        reply = self.send(mode='read', number=3, bits=bits)
        self.assertEqual(reply.parameter_value, 1000)
        self.assertEqual(reply.frequency, 1000)
        
        # TODO: Investigate whether there might be an error with these limits
        # in the pump manual.        
        # Set a new lower limit for the lower limit of the setpoint
        # (default = 2000).
        self.send(mode='write', number=20, value=0, bits=bits)
        # Set a new lower limit for the setpoint (default = 2000).
        self.send(mode='write', number=19, value=1000, bits=bits)
        # Set a new upper limit for the setpoint (default = 1000).
        self.send(mode='write', number=18, value=2000, bits=bits)
        # Set a new setpoint.
        reply = self.send(mode='write', number=24, value=1234, bits=bits)
        
        # Wait for the frequency to change.
        time.sleep(2 * self.vp.hardware_component.step)
        
        # Ask for frequency again.
        reply = self.send(mode='read', number=3, bits=bits)
        self.assertEqual(reply.parameter_value, 1234)
        self.assertEqual(reply.frequency, 1234)
        
        
if __name__ == '__main__':
    unittest.main()



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))

# The TurboCtl directory should already be in $PATH.

# -- Project information -----------------------------------------------------
import turboctl

project = 'TurboCtl'
author = 'Feliks Kivelä'
version = turboctl.__version__
copyright = turboctl.copyright
# The copyright variable is automatically written to the HTML footer.

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.

# Autodoc: Generate documentation from docstrings automatically.
# Napoleon: Support for Google-style docstrings.
# Intersphinx: Link to the documentaion of other sphinx projects.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',
    'sphinx.ext.intersphinx',
]

# Use the class docstring (instead of the __init__ docstring) for class
# descriptions.
# This seems to be the default value, but is specified just in case.
autoclass_content = 'class'

# Document classes, functions etc. in the same order as they are defined in the
# source code, instead of alphabetical order.
autodoc_member_order = 'bysource'

# This allows function signatures to be specified on the first line of the
# docstring.
# This is very useful for preventing sphinx from expanding DEFAULT in
# "func(arg=DEFAULT)".
autodoc_docstring_signature = True

autodoc_default_options = {
	# Print members (i.e. attributes, methods, classes etc.) when documenting
	# a module or an object.
	'members': True,
	# Print special members such as __init__ and __repr__.
	'special-members': True,
	# Print members in the same order as they as in the source files.
	'member-order': 'bysource',
	# :special-members: prints __weakref__ by default; this disables that.
    'exclude-members': '__weakref__',
}

# Projects linked to by Intersphinx.
intersphinx_mapping = {'python': ('https://docs.python.org/3', None),
					   'serial': ('https://pyserial.readthedocs.io/en/latest', None),
					   'urwid': ('http://urwid.org', None),}

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.

html_theme = 'alabaster'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".

html_static_path = ['_static']



"""Unit tests for the datatypes module."""

import math
import unittest

from hypothesis import given, assume, strategies as st

from turboctl.telegram.datatypes import (
    maxuint, maxsint, minsint, Uint, Sint, Float, Bin)


# Testing with large numbers provides no benefits and slows the tests
# down considerably.
MAX_BITS = 64
MAX_UINT = maxuint(MAX_BITS)
MAX_SINT = maxsint(MAX_BITS)
MIN_SINT = minsint(MAX_BITS)


### Hypothesis strategies ###

@st.composite
def values_and_bits(draw, class_, min_bits=0, max_bits=MAX_BITS):
    """Generate tuples of valid arguments for initializing Data
    objects.
    
    The type of *value* will always match the class; e.g.
    values_and_invalid_bits(Float) only generates floats and not ints or bytes
    objects.
    """
    bits = draw(st.integers(min_value=min_bits, max_value=max_bits))
    
    if class_ == Uint:
        value = draw(st.integers(min_value=0, max_value=maxuint(bits)))
    
    elif class_ == Sint:
        value = draw(st.integers(min_value=minsint(bits),
                                 max_value=maxsint(bits)))
    elif class_ == Float:
        bits = 32
        value = draw(st.floats(width=bits))
    
    elif class_ == Bin:
        value = draw(st.from_regex(f'\\A[01]{{{bits}}}\\Z'))
        
    return value, bits


@st.composite
def values_and_invalid_bits(draw, class_, min_bits=0, max_bits=MAX_BITS):
    """Generate tuples of a consisting of a *value* argument and an invalid
    *bits* argument for a given Data subclass.
    
    The type of *value* will always match the class; e.g.
    values_and_invalid_bits(Float) only generates floats and not ints or bytes
    objects.

    *bits* will be a non-negative int, but *value* will be too
    large or too small to be stored by that number of bits.
    """
    bits = draw(st.integers(min_value=min_bits, max_value=max_bits))    
    
    if class_ == Uint:
        # + 2 to make sure max_value > min_value.
        value = draw(st.integers(min_value=maxuint(bits) + 1, 
                                 max_value=MAX_UINT + 2))
    
    elif class_ == Sint:
        too_small = st.integers(min_value=MIN_SINT - 2,
                                max_value=minsint(bits) - 1)
        too_large = st.integers(min_value=maxsint(bits) + 1,
                                max_value=MAX_SINT + 2)
        value = draw(st.one_of(too_small, too_large))

    elif class_ == Float:
        # Bits values other than 32 are always invalid.
        bits_below_32 = st.integers(min_value=0, max_value=31)
        bits_above_32 = st.integers(min_value=33, max_value=MAX_BITS)
        bits = draw(st.one_of(bits_below_32, bits_above_32))
        value = draw(st.floats(width=32))
    
    elif class_ == Bin:
        length = draw(st.integers(min_value=bits + 1, max_value=MAX_BITS + 2))
        value = draw(st.from_regex(f'\\A[01]{{{length}}}\\Z'))

    return value, bits


@st.composite
def classes_values_and_invalid_bits(draw, min_bits=0, max_bits=MAX_BITS):
    """Like values_and_invalid_bits, but the class will also be randomly
    determined, and the returned value will be a tuple of (class, value, bits).
    """
    class_ = draw(data_types())
    value, bits = draw(values_and_invalid_bits(class_, min_bits, max_bits))
    return class_, value, bits


@st.composite
def data_types(draw, classes=(Uint, Sint, Float, Bin)):
    """Generate Data subclasses."""
    maxi = len(classes) - 1
    i = draw(st.integers(min_value=0, max_value=maxi))
    return classes[i]


@st.composite
def data_objects(draw, classes=(Uint, Sint, Float, Bin), **kwargs):
    """Generate instances of Data subclasses."""
    class_ = draw(data_types(classes))
    args = draw(values_and_bits(class_, **kwargs))
    return class_(*args)


@st.composite
def data_objects_and_slices(draw, **kwargs):
    """Generate references to all four Data subclasses."""
    obj = draw(data_objects(**kwargs))
    slice_ = draw(st.slices(obj.bits))
    return obj, slice_


class TestAttributes(unittest.TestCase):
    
    # value
    
    @given(data_objects())
    def test_value_is_readable_but_not_writable(self, obj):
        value = obj.value
        with self.assertRaises(AttributeError):
            obj.value = value
            
    @given(data_objects([Uint]))
    def test_uint_value_is_builtin_int(self, obj):
        self.assertIsInstance(obj.value, int)
        
    @given(data_objects([Sint]))
    def test_sint_value_is_builtin_int(self, obj):
        self.assertIsInstance(obj.value, int)
        
    @given(data_objects([Float]))
    def test_float_value_is_builtin_float(self, obj):
        self.assertIsInstance(obj.value, float)
        
    @given(data_objects([Bin]))
    def test_bin_value_is_binary_str(self, obj):
        self.assertIsInstance(obj.value, str)
        self.assertEqual(obj.value.strip('01'), '')
           
    # bits
            
    @given(data_objects())
    def test_bits_is_readable_but_not_writable(self, obj):
        bits = obj.value
        with self.assertRaises(AttributeError):
            obj.bits = bits
          
    @given(data_objects([Uint]))
    def test_uint_has_correct_bits(self, obj):
        value = obj.value
        bits = obj.bits
        self.assertTrue(0 <= value <= maxuint(bits))
        
    @given(data_objects([Sint]))
    def test_sint_has_correct_bits(self, obj):
        value = obj.value
        bits = obj.bits
        self.assertTrue(minsint(bits) <= value <= maxsint(bits))
        
    @given(data_objects([Float]))
    def test_float_has_32_bits(self, obj):
        self.assertEqual(obj.bits, 32)
        
    @given(data_objects([Bin]))
    def test_bin_has_correct_bits(self, obj):
        value = obj.value
        bits = obj.bits
        self.assertEqual(len(value), bits)
            
    # n_bytes

    @given(data_objects())
    def test_n_bytes_is_readable_but_not_writable(self, obj):
        n_bytes = obj.n_bytes
        with self.assertRaises(AttributeError):
            obj.n_bytes = n_bytes
            
    @given(data_objects())
    def test_padding_is_between_0_and_7_bits(self, obj):
        bits = obj.bits
        n_bytes = obj.n_bytes
        padding = 8*n_bytes - bits
        self.assertTrue(0 <= padding <= 7)
            

class TestMagicMethods(unittest.TestCase):
    
    # __bytes__
            
    @given(data_objects())
    def test_bytes_returns_a_bytes_object(self, obj):
        self.assertIsInstance(bytes(obj), bytes)
    
    @given(data_objects())
    def test_bytes_objects_have_a_len_of_n_bytes(self, obj):
        bytes_ = bytes(obj)
        self.assertEqual(len(bytes_), obj.n_bytes)
        
    @given(data_objects(),
            data_types(classes=(Uint, Sint, Bin)))
    def test_nested_bytes_when_class_is_not_float(self, obj, class_):
        bytes1 = bytes(obj)
        bytes2 = bytes(class_(bytes(obj)))
        self.assertEqual(bytes1, bytes2)
        
    @given(data_objects(min_bits=32, max_bits=32))
    def test_nested_bytes_when_class_is_float(self, obj):
        float_obj = Float(bytes(obj))
        # NaN has several different binary representations.
        assume(not math.isnan(float_obj.value))
        bytes1 = bytes(obj)
        bytes2 = bytes(float_obj)
        self.assertEqual(bytes1, bytes2)
        
    # __add__
        
    @given(data_objects(), data_objects())
    def test_bits_of_sum_is_sum_of_bits(self, obj1, obj2):
        sum_ = obj1 + obj2
        sum_of_bits = obj1.bits + obj2.bits
        self.assertEqual(sum_.bits, sum_of_bits)
        
    @given(data_objects(), data_objects())
    def test_add_concatenates_binary_representations(self, obj1, obj2):
        sum_ = obj1 + obj2
        sum_of_bins = Bin(obj1) + Bin(obj2)
        self.assertEqual(Bin(sum_), sum_of_bins)
                
    # __eq__
        
    @given(data_objects())
    def test_equality_is_reflexive(self, obj):
        self.assertEqual(obj, obj)
        
    @given(data_objects())
    def test_copy_is_equal_to_original(self, obj):        
        class_ = type(obj)
        other = class_(obj.value, obj.bits)
        self.assertEqual(obj, other)
        
    # __eq__ should work also against non-Data objects,
    # which is why this is tested with a few integers.
    @given(data_objects(), 
            st.one_of(data_objects(), 
                      st.integers(min_value=-10, max_value=10)))
    def test_different_type_value_or_bits_breaks_equality(self, obj1, obj2):
        try:
            both_values_nan = math.isnan(obj1.value) and math.isnan(obj2.value)
        except (TypeError, AttributeError):
            both_values_nan = False
        
        assume(
            type(obj1) != type(obj2)
            or ((obj1.value != obj2.value) and not both_values_nan)
            or obj1.bits != obj2.bits
        )
        self.assertNotEqual(obj1, obj2)
        
    # __repr__

    @given(data_objects())
    def test_eval_repr_returns_copy_except_for_inf_and_nan(self, obj):
        names = {**globals(), 'inf': float('inf'), 'nan': float('nan')}
        copy = eval(repr(obj), names)
        self.assertEqual(obj, copy)

    # __getitem__
        
    @given(data_objects([Bin]), st.slices(MAX_BITS))
    def test_getitem_slices_value_for_bin_objects(self,  obj, slice_):
        class_ = type(obj)
        self.assertEqual(obj[slice_], class_(obj.value[slice_]))
        
    @given(data_objects(), st.slices(MAX_BITS))
    def test_getitem_slices_binary_value(self, obj, slice_):
        self.assertEqual(Bin(obj[slice_]), Bin(obj)[slice_])
        

class TestInit(unittest.TestCase):
    
    @given(data_objects(), data_types())
    def test_type_conversions_preserve_bits_and_bytes(self, obj, class_):
        # Non-32 bit objects can't be cast to Float.
        if class_ == Float:
            assume(obj.bits==32)
        
        other = class_(obj)
        self.assertEqual(bytes(obj), bytes(other))
        self.assertEqual(obj.bits, other.bits)
    
    @given(values_and_bits(Uint))
    def test_uint_from_int_sets_attributes_correctly(self, i_and_bits):
        i, bits = i_and_bits        
        uint = Uint(i, bits)
        self.assertEqual(uint.value, i)
        self.assertEqual(uint.bits, bits)
        
    @given(values_and_bits(Sint))
    def test_sint_from_int_sets_attributes_correctly(self, i_and_bits):
        i, bits = i_and_bits
        sint = Sint(i, bits)
        self.assertEqual(sint.value, i)
        self.assertEqual(sint.bits, bits)
        
    @given(values_and_bits(Float))
    def test_float_from_float_sets_attributes_correctly(self, x_and_bits):
        x, bits = x_and_bits        
        float_ = Float(x, bits)
        
        if math.isnan(x):
            self.assertTrue(math.isnan(float_.value))
        else:
            self.assertEqual(float_.value, x)
        
        self.assertEqual(float_.bits, 32)
        
    @given(values_and_bits(Float))
    def test_float_from_int_sets_attributes_correctly(self, x_and_bits):
        x, bits = x_and_bits
        # x must be a numerical value so that it can be converted to int.
        assume(not math.isnan(x) and not math.isinf(x))
        i = int(x)        
        float_ = Float(i, bits)
        
        self.assertEqual(float_.value, i)
        self.assertEqual(float_.bits, 32)
        
    def test_float_sets_small_value_to_0(self):
        self.assertEqual(Float(1E-50).value, 0)
        self.assertEqual(Float(-1E-50).value, 0)
        
    @given(values_and_bits(Bin))
    def test_bin_from_str_sets_attributes_correctly(self, str_and_bits):
        s, bits = str_and_bits
        
        # Bits specified.
        bin_ = Bin(s, bits)
        self.assertEqual(bin_.value, s)        
        self.assertEqual(bin_.bits, bits)
        
        # Bits not specified.
        bin_ = Bin(s)
        self.assertEqual(bin_.value, s)        
        self.assertEqual(bin_.bits, len(s))
        
    @given(values_and_bits(Bin))
    def test_bin_from_int_sets_attributes_correctly(self, str_and_bits):
        s, bits = str_and_bits
        # '' is a valid 0-bit representation for 0. 
        i = int(s, 2) if s else 0

        bin_ = Bin(i, bits)
        self.assertEqual(bin_.value, s)        
        self.assertEqual(bin_.bits, bits)
                
    def test_bits_default_value(self):
        self.assertEqual(Uint(1).bits, 8)
        self.assertEqual(Sint(-1).bits, 8)
        self.assertEqual(Float(1.0).bits, 32)
        self.assertEqual(Float(1).bits, 32)
        self.assertEqual(Bin(1).bits, 8)
        # Bin doesn't have a default bit number when it's initialized from a
        # str.
        
        
class TestErrors(unittest.TestCase):
    
    ### Invalid value ###
    
    def test_uint_with_negative_value_raises_ValueError(self):
        with self.assertRaises(ValueError):
            Uint(-123)
            
    # Sint should work with all int values.
    
    def test_float_with_too_large_value_raises_ValueError(self):
        with self.assertRaises(ValueError):
            Float(1E50)
        with self.assertRaises(ValueError):
            Float(-1E50)
                        
    def test_bin_from_invalid_str_raises_ValueError(self):
        with self.assertRaises(ValueError):
            Bin('10201')
        
    ### Invalid bits ###
        
    # This only tests Floats with a float argument and Bins with a str
    # argument, so int arguments are tested separately below. 
    @given(classes_values_and_invalid_bits())
    def test_invalid_bits_raises_ValueError(self, class_value_and_bits):
        class_, value, bits = class_value_and_bits
        with self.assertRaises(ValueError):
            class_(value, bits)
 
    @given(values_and_invalid_bits(Float))
    def test_invalid_bits_raises_ValueError_for_float_from_int(self,
                                                               x_and_bits):
        x, bits = x_and_bits
        # x must be a numerical value so that it can be converted to int.
        assume(not math.isnan(x) and not math.isinf(x))
        i = int(x)        
        
        with self.assertRaises(ValueError):
            Float(i, bits)
            
    # We can simply generate Uint values, since the range of valid int
    # arguments is the same for both Uints and Bins.
    @given(values_and_invalid_bits(Uint))
    def test_invalid_bits_raises_ValueError_for_bin_from_int(self, i_and_bits):
        i, bits = i_and_bits
        
        with self.assertRaises(ValueError):
            Bin(i, bits)
            
    # The tests above only test positive *bits* values, so negative values need
    # to be tested separately.
    # Floats and Bins with int arguments are again tested in different methods.
    @given(classes_values_and_invalid_bits(),
           st.integers(min_value=-MAX_BITS, max_value=-1))
    def test_negative_bits_raises_ValueError(self, class_value_and_bits, 
                                             negative_bits):     
        class_, value, _ = class_value_and_bits
        with self.assertRaises(ValueError):
            class_(value, negative_bits)
            
    @given(values_and_invalid_bits(Float),
           st.integers(min_value=-MAX_BITS, max_value=-1))
    def test_negative_bits_raises_ValueError_for_float_from_int(
            self, x_and_bits, negative_bits):
        
        x, _ = x_and_bits
        assume(not math.isnan(x) and not math.isinf(x))
        i = int(x)        
        
        with self.assertRaises(ValueError):
            Float(i, negative_bits)
            
    @given(values_and_invalid_bits(Uint),
           st.integers(min_value=-MAX_BITS, max_value=-1))
    def test_negative_bits_raises_ValueError_for_bin_from_int(
            self, i_and_bits, negative_bits):
        
        i, _ = i_and_bits
        
        with self.assertRaises(ValueError):
            Bin(i, negative_bits)
        
    
    ### Bits specified when it shouldn't be ###
    
    @given(data_objects())
    def test_bits_arg_raises_TypeError_when_initializing_from_data_obj(self, 
                                                                       obj):
        class_ = type(obj)
        with self.assertRaises(TypeError):
            class_(obj, obj.bits)
            
    @given(data_objects())
    def test_bits_arg_raises_TypeError_when_initializing_from_bytes(self, obj):
        class_ = type(obj)
        with self.assertRaises(TypeError):
            class_(bytes(obj), obj.bits)


if __name__ == '__main__':
    unittest.main()



"""Write README.rst with and without Sphinx extensions."""
import turboctl
from turboctl.ui.docs import URL


# This is the main text of the readme file.
# The documentation section as well as the current version and copyright
# information need to be filled in.
body = """.. Note: This file is automatically generated by make_readme.py.
.. To make changes to this file, edit that instead of this.

TurboCtl
========

TurboCtl is a RS-232 controller for a Leybold TURBOVAC i or iX turbovacuum
pump. It was written in Python 3 and is intended for Linux operating systems.

Documentation
-------------

{documentation}

Author
------

TurboCtl was written by Feliks Kivelä.
The author can be contacted at firstname.lastname@aalto.fi
(with the "ä" replaced by an "a").

Copyright © {project_copyright}.

GitHub
------

The GitHub home page of the TurboCtl project can be found
`here <https://github.com/fkivela/TurboCtl>`__.

Version
-------

This is version {version} of TurboCtl.

License
-------

TurboCtl is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see https://www.gnu.org/licenses/.
"""
# Note the double underscores at the ends of links.
# With a regular single underscore the duplicate link names won't be accepted
# by PyPI.


# This version displays correctly in GitHub and PyPI, which don't support
# Sphinx.
raw_rst_docs = f'You can view the TurboCtl documentation `here <{URL}>`__.'


# This version is used for the index page of the HTML docs.
sphinx_docs = """.. toctree::

   installation.rst
   usage.rst
   modules.rst
   errata.rst
   changelog.rst"""


# Write the readme file.
with open('README.rst', 'w') as file:
    file.write(body.format(documentation=raw_rst_docs,
                           project_copyright=turboctl.copyright,
                           version=turboctl.__version__))


# Write the HTML docs index.
with open('doc/sphinx/index.rst', 'w') as file:
    file.write(body.format(documentation=sphinx_docs,
                           project_copyright=turboctl.copyright,
                           version=turboctl.__version__))



"""This module is used to create formatted tables of parameters, errors and
warnings.
"""

import textwrap as tx
import tabulate

from turboctl.telegram.datatypes import Data


tabulate.PRESERVE_WHITESPACE = True
        

def table(database, numbers='all', widths={}):
    """Return a table of parameters, errors or warnings.
    
    Args:
        database: A :class:`dict` of parameters, errors or warnings. The values 
            should be :class:`~turboctl.telegram.parser.Parameter` or 
            :class:`~turboctl.telegram.parser.ErrorOrWarning` objects, and the
            keys the numbers (:class:`int`) of those objects.
        numbers: ``'all'``, or a sequence of the numbers of the 
            parameters/errors/warnings that should be displayed.
        widths: A :class:`dict` with keys corresponding to column names and 
            values to the maximum widths of the columns. The column names 
            are the same as the attributes of the
            :class:`~turboctl.telegram.parser.Parameter` or 
            :class:`~turboctl.telegram.parser.ErrorOrWarning` classes.
            If the maximum width of a column isn't specified in *widths*, it
            is set to infinite and no text wrapping is used for that column.
    
    Returns:
        The table as a formatted :class:`str` ready for printing. 
            
    Raises:
        ValueError: If *database* is empty or if there is no 
            parameter/error/warning corresponding to a number 
            in *numbers*.
    """
    a = array(database, numbers, widths)
    return tabulate.tabulate(a, tablefmt='plain')

def array(database, numbers='all', widths={}):
    """The same as :func:`table`, but instead of a :class:`str` the table is
    returned as a :class:`list` of rows, each of which is a :class:`list` of
    fields (:class:`str`).
    
    This function can be used to test this module.
    """
    if numbers == 'all':
        data_objects = database.values()
        
    else:
        data_objects = []
        for n in numbers:
            try:
                data_objects.append(database[n])
            except KeyError:
                raise ValueError(f'Invalid number: {n}')

    # If a maximum width is not specified, it is set to infinite.
    for f in _column_names(database):
        if f not in widths:
            widths[f] = float('inf')
    
    return [_format_row(o, widths) for o in data_objects]
        
def _column_names(database):
    """Return the names of the columns in the table.
    
    The column names correspond to the attributes of the objects in 
    *database*.
    It is assumed *database* only contains objects of one type 
    (Parameter or ErrorOrWarning).
    
    Returns:
        A list of strings.
        
    Raises:
        ValueError: If *database* is empty.
    """
    try:
        object_ = list(database.values())[0]
    except KeyError as e:
        raise ValueError('*database* is empty') from e
        
    return list(object_.fields.keys())

def _format_row(data_object, widths):
    """Return a single table row.
    
    Args:
        data_object: The object whose data should be displayed on 
            the row.
        widths: The maximum widths of the cells in the row (a list).
        
    Returns: A list of strings corresponding to the text displayed in
        each cell of the row.
    """    
    return [_format_field(name, value, widths[name])
            for name, value in data_object.fields.items()]

def _format_field(name, value, width):
    """Return a single formatted table cell.
    
    Args:
        name: The name of the column the cell belongs to.
        value: The value to be displayed in the cell (any type).
        width: The maximum width of the cell (in characters).
        
    Returns: A string. Line breaks will be inserted in appropriate 
        places, if some of the lines in text would be longer than 
        *width* wthout them. If *value* already contains line breaks, 
        these are preserved.
    """
    # Format fields differently based on the type of the value.
    if isinstance(value, str):
        # Replace empty text fields with a dash. 
        if value == '':
            value = '-'        
        # Add line breaks to text fields.
        value = _wrap(value, width)
    
    elif isinstance(value, range):
        # Display the 'indices' field in a nicer format instead of
        # "range(i, j)".
        value = f'{value.start}...{value.stop-1}' if value else '-'            
    
    elif isinstance(value, Data):
        # Convert Data subclass instances (i.e. the values of the the
        # 'min_value', 'max_value' and 'default' fields) to built-in types.
        value = value.value
        
        # Truncate long floats.
        if isinstance(value, float):
            value = f'{value:.7}'
    
    elif isinstance(value, list):
        # The 'default' field may contain a list of Data subclass instances,
        # which need to be converted to built-in types.
        # list is a mutable type, so we need to make a copy in order to avoid
        # making changes to the parameter itself. 
        value = value.copy()
        for i, item in enumerate(value):
            value[i] = item.value
            
    elif isinstance(value, type):
        # Display the name of classes.
        value = value.__name__
    
    # Replace underscores in the field name with spaces.
    # E.g. 'min_value' -> 'min value'         
    name = name.replace('_', ' ')
    
    # The field name is written in all capitals.
    # Tabulate seems to strip a '\n' from the end of a string, 
    # so two line breaks are needed to add an empty line between rows.
    return name.upper() + '\n' + str(value) + '\n\n'

def _wrap(string, width):
    """Add line breaks to a long string.
    
    Args:
        string: The string that line breaks should be insterted to.
        width: The maximum amount of characters in each line.
    
    Returns:
        A string with line breaks inserted in appropriate places. 
        If *string* already contains line breaks, these are 
        preserved.
    """
    lines = string.split('\n')
    wrapped_lines = [tx.wrap(line, width) for line in lines]
    shortened_lines = ['\n'.join(list_) for list_ in wrapped_lines]
    return '\n'.join(shortened_lines)



"""Unit tests for the api module."""

import unittest

import serial

from turboctl.telegram import api
from turboctl.telegram.codes import ControlBits
from turboctl.virtualpump.virtualpump import VirtualPump


class TestApi(unittest.TestCase):
    
    def setUp(self):
        self.vp = VirtualPump()
        self.connection = serial.Serial(self.vp.connection.port, timeout=1)
        
    def tearDown(self):
        self.vp.stop()
        
    def test_test(self):
        self.assertTrue(True)
        
    def check_parameter(self, telegram, type_, mode, number, value, index):
        self.assertEqual(telegram.type, type_)
        self.assertEqual(telegram.parameter_mode, mode)
        self.assertEqual(telegram.parameter_number, number)
        self.assertEqual(telegram.parameter_value, value)
        self.assertEqual(telegram.parameter_index, index)
        
    def check_hardware(self, telegram, frequency, temperature, current,
                       voltage, bits):
        self.assertEqual(telegram.type, 'query')
        self.assertEqual(telegram.frequency, 0)
        self.assertEqual(telegram.temperature, 0)
        self.assertEqual(telegram.current, 0)
        self.assertEqual(telegram.voltage, 0)
        # Compare sets so that the order doesn't matter.
        self.assertEqual(set(telegram.flag_bits), set(bits))
        
    def test_write_and_read_unindexed_parameter(self):
        # Parameter 16 has the following relevant properties:
        # min=0 max=150 default=80 access=r/w type=s16.
        
        # Write a value to the parameter.
        query, reply = api.write_parameter(self.connection, 16, 123)
        self.check_parameter(query, 'query', 'write', 16, 123, 0)
        self.check_parameter(reply, 'reply', 'response', 16, 123, 0)
        
        # Read the written value.
        query, reply = api.read_parameter(self.connection, 16)
        self.check_parameter(query, 'query', 'read', 16, 0, 0)
        self.check_parameter(reply, 'reply', 'response', 16, 123, 0)
        
    def test_write_and_read_indexed_parameter(self):
        # Parameter 26 has the following relevant properties:
        # min=0 max=65535 default=25 access=r/w type=u16 indices=[0,1,2]
        
        # Write a value to the parameter.
        query, reply = api.write_parameter(self.connection, 26, 123, 1)
        self.check_parameter(query, 'query', 'write', 26, 123, 1)
        self.check_parameter(reply, 'reply', 'response', 26, 123, 1)
        
        # Read the written value.
        query, reply = api.read_parameter(self.connection, 26, 1)
        self.check_parameter(query, 'query', 'read', 26, 0, 1)
        self.check_parameter(reply, 'reply', 'response', 26, 123, 1)
        
    def test_onoff_cycle(self):
        
        on_bits = [ControlBits.COMMAND, ControlBits.ON]
        
        # There's no need to properly test reply, since the data it contains is
        # determined by VirtualPump, which is tested separately.
        # It is enough to make sure the reply exists and actually represents a
        # reply.
        
        # Ask pump status when the pump is off.
        query, reply = api.status(self.connection, pump_on=False)
        self.check_hardware(query, 0, 0, 0, 0, [])
        self.assertEqual(reply.type, 'reply')
        self.assertFalse(self.vp.hardware_component.is_on)
        
        # Turn the pump on.
        query, reply = api.status(self.connection, pump_on=True)
        self.check_hardware(query, 0, 0, 0, 0, on_bits)
        self.assertEqual(reply.type, 'reply')
        self.assertTrue(self.vp.hardware_component.is_on)
        
        # Turn the pump off again.
        query, reply = api.status(self.connection, pump_on=False)
        self.check_hardware(query, 0, 0, 0, 0, [])
        self.assertEqual(reply.type, 'reply')
        self.assertFalse(self.vp.hardware_component.is_on)


if __name__ == '__main__':
    unittest.main()



from datetime import date


__version__ = '1.1.1.dev3'
"""The current version of the TurboCtl project."""


copyright = f'2019-{date.today().year} University of Helsinki Fusor Team'
"""Automatically updating copyright information."""



"""This module defines the :class:`AdvancedTUI` class.

:class:`AdvancedTUI` uses `urwid <http://urwid.org/>`_.
If urwid hasn't been installed,
:class:`~turboctl.ui.command_line_ui.CommandLineUI` should be used instead.
"""
import threading
import urwid

from turboctl.ui import widgets


class AdvancedTUI(urwid.WidgetWrap):
    """A text-based user interface combining a terminal-style
    command line interface and a text screen.

    Attributes:
        display (:class:`urwid.Text`):
            A text field for presenting information to the user.
            Use :attr:`display`:meth:`.set_text()
            <urwid.Text.set_text>` to change its contents.

        command_line_interface(:class:`~turboctl.ui.widgets\
.ScrollableCommandLines`):
            A command-line interface for issuing commands, located
            below :attr:`display`.
    """

    def __init__(self, script, inputfile, outputfile, palette=()):
        """Initialize a new :class:`AdvancedTUI`.

        Args:
            script:
                The program that this UI controls.
                It should be wrapped in a callable object so that
                calling ``script()`` executes it.
                The UI communicates with the program through
                *inputfile* and *outputfile*;
                the program should read its input from *inputfile*
                and send its output to *outputfile*.
            inputfile (file-like object):
                User input entered into the UI is written into this
                object.
            outputfile (file-like object):
                Output printed on the screen by the UI is read from
                this object.
            palette (iterable of palette entries):
                The palette used for the UI.
                This is passed to the initializer of
                :class:`urwid.MainLoop`.
                The format used by palette entries is specified in the
                documentation of
                :meth:`urwid.BaseScreen.register_palette`.
        """
        self.display = urwid.Text('')
        self.command_line_interface = widgets.ScrollableCommandLines(
            inputfile, outputfile)
        self._stop_flag = threading.Event()

        upper_half = self.display
        divider = urwid.Divider('─')
        lower_half = self.command_line_interface
        # 'pack' treats upper_half and divider as flow widgets and
        # prevents the adding of padding to them.
        screen = urwid.Pile(
            [('pack', upper_half), ('pack', divider), lower_half])
        super().__init__(screen)

        def run_script():
            script()
            self._stop_flag.set()

        self._thread = threading.Thread(target=run_script, daemon=True)
        self._loop = urwid.MainLoop(self, palette)
        self._loop.set_alarm_in(0.01, self._callback, user_data=None)

    def _callback(self, loop, user_data):
        """Update the UI periodically.

        This method makes self.command_line_interface.command_lines
        handle new output and re-renders
        self.command_line_interface.scrollbar,
        since printing new output may change its size or position.

        If the program has finished its execution, this method
        breaks the UI loop.

        A call to this method sets an alarm which calls it again after
        a set period; this repeats until the UI loop is broken.
        """
        if self._stop_flag.is_set():
            raise urwid.ExitMainLoop

        self.command_line_interface.command_lines.update()
        # _invalidate marks a widget for re-rendering.
        # urwid documentation suggests using this method even though
        # it begins with '_'.
        # pylint: disable=protected-access
        self.command_line_interface.scrollbar._invalidate()
        # pylint: enable=protected-access

        self._loop.set_alarm_in(0.01, self._callback, user_data=None)

    def run(self):
        """Run the program and the UI loop in parallel threads.
        The UI loop beaks automatically when the program ends.

        .. Warning::
            If the UI loop (which is executed in the main thread)
            crashes,
            the program (executed in a parallel thread) will probably
            be left waiting for input in a blocking state.
            Safeguards to prevent this should be built into code that
            uses :class:`AdvancedTUI`; for example, this method may
            be called in a ``with`` or a ``try-finally`` block that
            makes sure the program is closed when the block is
            exited.
        """
        self._thread.start()
        self._loop.run()


"""Functionality for opening TurboCtl documentation in a browser."""
import webbrowser


URL = 'https://turboctl.readthedocs.io/en/latest/index.html'
"""The URL of TurboCtl documentation."""


def docs():
    """Open TurboCtl documentation in a web browser."""
    webbrowser.open(URL, new=2)



"""This module defines classes for creating, representing and reading
telegrams which are used to communicate with the pump.
    
..
    Aliases for Sphinx.

.. |Uint| replace:: :class:`~turboctl.telegram.datatypes.Uint`
.. |Sint| replace:: :class:`~turboctl.telegram.datatypes.Sint`
.. |Float| replace:: :class:`~turboctl.telegram.datatypes.Float`
.. |Bin| replace:: :class:`~turboctl.telegram.datatypes.Bin`
"""

from dataclasses import dataclass

from turboctl.telegram.codes import (
    ControlBits, StatusBits, get_parameter_code, get_parameter_mode,
    ParameterResponse, ParameterError
)
from turboctl.telegram.datatypes import (Data, Uint, Sint, Bin)
from turboctl.telegram.parser import PARAMETERS


@dataclass
class Telegram:
    """A simple dataclass that represents a telegram sent to or from the pump.
    
    This class is cumbersome to initialize directly, since the values of all 
    attributes must be given as arguments. Instances should instead be 
    created with the :class:`TelegramBuilder` class.
    
    The Leybold TURBOVAC i/iX vacuum pump communicates with a computer 
    via its RS 232 or RS 485 serial port or a USB port using telegrams 
    of 24 bytes. The general structure of the telegrams follows the 
    USS protocol.
    
    Each byte consists of a start bit (0), 8 data bits, an even 
    parity bit (1 if there are an even number of 1s in the data 
    bits, 0 otherwise) and an ending bit (1). However, only the data 
    bits are included in the bytes objects that represent telegrams;
    the `serial <https://pyserial.readthedocs.io/en/latest/>`_ 
    module automatically adds the other bits. 
    
    In the TURBOVAC manual, the data bits in a byte are indexed as 
    [7,6,5,4,3,2,1,0] (i.e. according to the power of 2 they
    represent), but this class uses the convention [0,1,2,3,4,5,6,7], 
    because it corresponds to the indices of a Python list.
    
    The functions and values of the different bytes in a telegram 
    are detailed below. Each list entry contains the name of the  
    :class:`Telegram` object attribute (if any) that represents 
    the value associated with the entry.
    
    Unless otherwise noted, all bytes have a default value of 0.
        
    **Byte 0:** STX (start of text). Always 2.
    
    **Byte 1:** LGE (telegram length). 
    Excludes bytes 0 and 1, and thus always has a value of 22.
    
    **Byte 2:** ADR (address). 
    With a RS485 port this denotes the slave node number (0-31).
    Reading or changing this byte is not currently supported by this 
    class.
    
    **Bytes 3-4:** PKE (parameter number and type of access). 
    A 16 bit block:
        
        - Bits 0-3: 
            Type of parameter access or response. 
            This is a 4 bit code indicating e.g. whether a parameter 
            should be read from or written to.
            Valid codes are detailed in the :mod:`~turboctl.telegram.codes` 
            module.
            
            Attribute: :attr:`parameter_code`.
            
        - Bit 4: 
            Always 0.
        
        - Bits 5-15:
            The number of the parameter to be accessed.
            
            Attribute: :attr:`parameter_number`.
    
    **Byte 5:** - (reserved). Always 0.
    
    **Byte 6:** IND (parameter index). 
    If the requested parameter is indexed, this specifies the number 
    of the requested index.
    
    Attribute: :attr:`parameter_index`.

    **Bytes 7-10:** PWE (parameter value). 
    This block contains a parameter value that is written to or read 
    from the pump. If the pump tries to access a parameter but fails, the 
    reply will contain an error code in this block.
    
    Attribute: :attr:`parameter_value`.

    **Bytes 11-12:** PZD1 (status and control bits).
    16 bits each corresponding to a single setting or command which 
    can be turned on by setting the bit to 1.
    In a reply from the pump these correspond to status 
    conditions affecting the pump instead of commands.
    
    Attribute: :attr:`flag_bits`.

    **Bytes 13-14:** PZD2 (current rotor frequency). 
    Rotor frequency in Hz; the same as parameter 3.
    Included in all replies, and can be included in queries to define 
    a setpoint for the frequency. (This only works if the setpoint is 
    enabled through the control bits, and overrides the setpoint 
    defined in parameter 24).
    
    Attribute: :attr:`frequency`.

    **Bytes 15-16:** PZD3 (current frequency converter temperature). 
    Frequency converter temperature in °C, included in all replies. 
    Same as parameter 11.
    
    Attribute: :attr:`temperature`.
    
    **Bytes 17-18**: PZD4 (current motor current). 
    Motor current in 0.1 A, included in all replies. 
    Same as parameter 5.
    
    Attribute: :attr:`current`.

    **Bytes 19-20:** - (reserved). Always 0.

    **Bytes 21-22:** PZD6 (current intermediate circuit voltage).
    Intermediate circuit voltage in V, included in all replies.
    Same as parameter 4.
    
    Attribute: :attr:`voltage`

    **Byte 23:** BCC (byte block check): 
    A checksum computed using the following algorithm:
    ::
        
        checksum = bytes_[0]
        for byte in bytes_[1:23]:
            checksum = checksum ^ byte_
                
    where ``^`` is the exclusive or (XOR) operator.    
    """
    # The manual uses "stator frequency (=P3)" instead of "rotor frequency",
    # but the entry for parameter 3 uses the word "rotor", which seems to be
    # the correct version since the rotor is the moving part.
    # The manual also lists the unit of voltage as 0.1 V, but the correct unit
    # seems to be V.
    
    # :mod:`serial` links to a weird place in pySerial's documentation, so a
    # manual hyperlonk to the front page was used instead.
    
    parameter_code: Bin
    """The parameter access or response code as a 4-bit |Bin|."""
    
    parameter_number: Uint
    """The parameter number as an 11-bit |Uint|."""
    
    parameter_index: Uint
    """The parameter index as a 8-bit |Uint|."""
    
    parameter_value: Data
    """The parameter value. This attribute is always a 32-bit instance of a 
    subclass of :class:`~turboctl.telegram.datatypes.Data`, 
    but the exact type depends on the parameter.
    """
    
    flag_bits: Bin
    """The control or status bits as a 32-bit |Bin|."""
    
    frequency: Uint
    """The frequency as a 32-bit |Uint|."""
    
    temperature: Sint
    """The temperature as a 32-bit |Sint|."""
    
    current: Uint
    """The current as a 32-bit |Uint|."""
    
    voltage: Uint
    """The voltage as a 32-bit |Uint|."""
    
    LENGTH = 24
    """The length of a telegram in bytes."""
    
    def __bytes__(self):
        """Return the telegram as a :class:`bytes` object.
        
        The checksum is computed automatically and added to the end.
        """
        bytes_ = bytes(
            Uint(2, 8) +
            Uint(22, 8) +
            Uint(0, 8) +
            self.parameter_code + 
            Bin('0') + 
            self.parameter_number + 
            Uint(0, 8) + 
            self.parameter_index + 
            self.parameter_value +
            self.flag_bits[::-1] +    
            self.frequency +
            self.temperature +
            self.current + 
            Uint(0, 16) +
            self.voltage
        )
        return bytes_ + bytes([checksum(bytes_)])


class TelegramBuilder:
    """
    TelegramBuilder(parameters=PARAMETERS)
    
    This class can be used to easily construct instances of the 
    :class:`Telegram` class.
    
    Here is an example of how to use this class:     
    ::
        
        telegram = (TelegramBuilder().set_parameter_mode('write')
                                     .set_paramerer_number(1)
                                     .set_parameter_index(2)
                                     .set_parameter_value(3)
                                     .build())
        
    The above creates a telegram which writes the value 3 to parameter 1,
    index 2. Note that this is just an example of the syntax; parameter 1
    isn't actually indexed.
   
    Attributes which aren't explicitly set to a value with a setter method are 
    set to zero when the telegram is created.
    Trying to set an attribute to an invalid value results in a
    :class:`ValueError` or a :class:`TypeError`.
        
    A telegram can also be created from a :class:`bytes` object:
    ::
        
        telegram = TelegramBuilder().from_bytes(bytes_).build()
        
    Attributes:
        parameters: A :class:`dict` of
            :class:`~turboctl.telegram.parser.Parameter` objects, with
            parameter numbers as keys.
            The default value is :const:`~turboctl.telegram.parser.PARAMETERS`,
            but non-default parameter sets can be used for testing purposes.
    """
    # The first line of the docstring overrides the default signature generated
    # by Sphinx, and thus prevents PARAMETERS from being expanded.
    
    def __init__(self, parameters=PARAMETERS):
        """
        __init__(parameters=PARAMETERS)
        
        Initialize a new :class:`TelegramBuilder`.
        
        Args:
            parameters: The object to be assigned to :attr:`parameters`.
        """
        self.parameters = parameters
        
        # Keyword arguments used to create a telegram.
        self._kwargs = {
            'parameter_code':   Bin(4 * '0', bits=4),
            'parameter_number': Uint(0,  bits=11),
            'parameter_index':  Uint(0,  bits=8),
            'parameter_value':  Uint(0,  bits=32),
            'flag_bits':        Bin(16 * '0', bits=16),
            'frequency':        Uint(0,  bits=16),
            'temperature':      Sint(0,  bits=16),
            'current':          Uint(0,  bits=16,),
            'voltage':          Uint(0,  bits=16),
        }
        # parameter_value and parameter_mode are special cases.
        # These variables store the values given by the user, and the final
        # values used as arguments are only determined when the telegram is
        # created.
        
        # __init__ and set_parameter_value set this to a int or a float,
        # and from_bytes to a bytes object.
        self._parameter_value = 0
        
        # __init__ and set_parameter_mode set this to a string, and from_bytes
        # to a Bin object.
        self._parameter_mode = 'none'

    def from_bytes(self, bytes_):
        """Read the contents of the telegram from a :class:`bytes` object.
        
        The type of :attr:`parameter_value <Telegram.parameter_value>`
        depends on :attr:`parameter_number <Telegram.parameter_number>`,
        and is assigned automatically. If *bytes_* contains a parameter 
        number that doesn't exist, a :class:`ValueError` is raised.
        
        If the parameter isn't accessed (i.e. the parameter mode is set to 
        ``'none'`` or code to ``'0000'``), invalid parameter numbers, such as
        the default value of 0, are permitted.
        In that case, the parameter type is set to
        :class:`~turboctl.telegram.datatypes.Uint`.
        
        Note that this isn't a class method; a :class:`TelegramBuilder` must
        first be initialized normally with :meth:`__init__`, after which
        this method may be called.
        
        Raises:
            ValueError: If *bytes_* doesn't represent a valid telegram.
        """
        self._check_valid_telegram(bytes_)
        
        code_and_number_bits = Bin(bytes_[3:5])    
        self._kwargs = {
            'parameter_number':     Uint(code_and_number_bits[5:16]),
            'parameter_index':      Uint(bytes_[6]),
            'flag_bits':            Bin(bytes_[11:13])[::-1],
            'frequency':            Uint(bytes_[13:15]),
            'temperature':          Sint(bytes_[15:17]),
            'current':              Uint(bytes_[17:19]),
            'voltage':              Uint(bytes_[21:23])        
        }
        
        self._parameter_value = bytes_[7:11]
        self._parameter_mode = Bin(code_and_number_bits[0:4])

        return self
        
    @staticmethod
    def _check_valid_telegram(bytes_):
        """Raise a ValueError if bytes_ doesn't represent a valid Telegram.
        """
        
        if len(bytes_) != 24:
            raise ValueError(f'len(bytes_) should be 24, not {len(bytes_)}')
        
        if bytes_[0] != 2:
            raise ValueError(f'bytes_[0] should be 2, not {bytes_[0]}')
            
        if bytes_[1] != 22:
            raise ValueError(f'bytes_[1] should be 22, not {bytes_[1]}')
            
        cs = checksum(bytes_[0:23])
        if cs != bytes_[23]:
            raise ValueError(f'bytes_[23] (the checksum) should be '
                             f'{cs}, not {bytes_[23]}')
    
    def set_parameter_mode(self, value: str):
        """Set the parameter access or response mode to one of the following:
            
            Access modes:
                - ``'none'``
                - ``'read'``
                - ``'write'``
                
            Response modes:
                - ``'none'``
                - ``'response'``
                - ``'error'``
                - ``'no write'``
                
            The parameter access or response code is determined automatically
            based on the parameter mode and the parameter number. 
            """
        self._parameter_mode = value
        return self
    
    def set_parameter_number(self, value: int):
        """Set the parameter number.
        
        Raises:
            ValueError: If there isn't a parameter with the specified number.
        """
        self._kwargs['parameter_number'] = Uint(value, 11)
        return self
        
    def set_parameter_index(self, value: int):
        """Set the parameter index."""
        self._kwargs['parameter_index'] = Uint(value, bits=8)
        return self
    
    def set_parameter_value(self, value):
        """Set the parameter value.
        
        The type of *value* depends on the type of the parameter.
        This method can also be used to set the error code; if
        :meth:`set_parameter_mode` is called to set the parameter mode to
        ``'error'``, the parameter value is always interpreted as an |Uint|
        error code regardless of parameter number or type.
        """
        self._parameter_value = value
        return self
        
    def set_flag_bits(self, bits):
        """Set the control or status bits.
        
        *bits* should be an iterable of those 
        :class:`~turboctl.telegram.codes.ControlBits` or 
        :class:`~turboctl.telegram.codes.StatusBits` members that should be 
        included in the telegram.
        """
        bitlist = 16 * ['0']
        for bit in bits:
            bitlist[bit.value] = '1'
        string = ''.join(bitlist)
        self._kwargs['flag_bits'] = Bin(string, bits=16)
        return self
    
    def set_frequency(self, value: int):
        """Set the frequency."""
        self._kwargs['frequency'] = Uint(value, bits=16)
        return self
    
    def set_temperature(self, value: int):
        """Set the temperature.
        
        Note that *value* can also be negative.
        """
        self._kwargs['temperature'] = Sint(value, bits=16)
        return self
    
    def set_current(self, value):
        """Set the current."""
        self._kwargs['current'] = Uint(value, bits=16)
        return self
    
    def set_voltage(self, value):
        """Set the voltage."""
        self._kwargs['voltage'] = Uint(value, bits=16)
        return self
        
    def build(self, type_='query'):
        """Build and return a :class:`Telegram` object.
        
        Args:
            type_: ``query`` if the telegram represents a message to the pump, 
                ``reply`` if it represents a message from the pump.
        
        Raises:
            :class:`ValueError` or :class:`TypeError`: If a telegram cannot
                be created with the specified attributes.
        """
        
        # Make sure type_ is valid to avoid bugs caused by a misspelled
        # argument.
        if type_ not in ['query', 'reply']:
            raise ValueError(f'invalid type_: {type_}')
        
        # Determine parameter access code.
        
        none_code = '0000'
        error_code = ParameterResponse.ERROR.value
        
        # __init__() and set_parameter_mode set self._parameter_mode to a
        # string, while from_bytes() sets it to a Bin object.
        mode_is_none = self._parameter_mode in ['none', Bin(none_code)]
        mode_is_error = (type_ == 'reply' and
                      self._parameter_mode in ['error', Bin(error_code)])
        
        if mode_is_none:
            self._kwargs['parameter_code'] = Bin(none_code)
        elif mode_is_error:
            self._kwargs['parameter_code'] = Bin(error_code)
        else:
            # parameter_number must be valid if the access mode isn't 'none'
            # or 'error'.
            number = self._kwargs['parameter_number'].value
            try:
                parameter = self.parameters[number]
            except KeyError:
                raise ValueError(f'invalid parameter number: {number}') 

            if isinstance(self._parameter_mode, Bin):
                # If this object was created from a Bin object,
                # self._parameter_mode will be a bytes object.
                self._kwargs['parameter_code'] = self._parameter_mode
            else:
                code = get_parameter_code(
                    type_,  self._parameter_mode,
                    bool(parameter.indices), parameter.bits
                ).value
                self._kwargs['parameter_code'] = Bin(code)

        # Determine parameter value.
        
        if mode_is_none or mode_is_error:
            # If the mode is 'none', there is no parameter access and the
            # datatype doesn't matter.
            # If the mode is 'error', the parameter value is replaced by an
            # Uint error code.
            datatype = Uint
        else:
            datatype = parameter.datatype
                
        if isinstance(self._parameter_value, bytes):
            # If self._parameter_value was set by from_bytes(),
            # self._parameter_value will be a bytes object and bits cannot be
            # specified.
            self._kwargs['parameter_value'] = datatype(self._parameter_value)
        else:
            # If _parameter_value was specified by the user, its type should
            # match the type of the parameter.
            # The default value of 0 is accepted by all parameter types.
            self._kwargs['parameter_value'] = datatype(self._parameter_value,
                                                       bits=32)

        return Telegram(**self._kwargs)


class TelegramReader:
    """This class can be used to read the data of a telegram in a more
    user-friendly way. This means the returned values are Python built-ins 
    instead of the custom datatypes used by the :class:`Telegram` class, and 
    *parameter_code* is automatically converted to a human-readable value.
        
    Attributes:        
        type:
            ``'query'`` for telegrams to the pump, ``'reply'`` for telegrams
            from the pump.
            
        telegram:
            The :class:`Telegram` object that is read.
    """
    
    def __init__(self, telegram, type_='reply'):
        """Initialize a new :class:`TelegramReader`.
        
        Args:
            telegram: The :class:`Telegram` to be read.
            type_: ``query`` if the telegram represents a message to the pump, 
                ``reply`` if it represents a message from the pump.
                
        Raises:
            :class:`ValueError`: If *type_* has an invalid value.
        """
        
        # Make sure type_ is valid to avoid bugs caused by a misspelled
        # argument.
        if type_ not in ['query', 'reply']:
            raise ValueError(f'invalid type_: {type_}')
        
        self.type = type_
        self.telegram = telegram
        
    def __repr__(self):
        """Return an exact string respresentation of this object.
        
        The returned string can be evaluated to create a copy of this object.
        The format is ``ClassName(telegram=<telegram>, type=<type>)``.
        """
        return type(self).__name__ + (
            f'(telegram={self.telegram}, type={repr(self.type)})')

        
    def __str__(self):
        """Return an easily readable string representation of this object.
        
        The returned string shows the values of both the attributes and the
        read-only properties of this object, and cannot thus be passed to
        :func:`eval` without an raising error. 
        
        The format is
        
        .. highlight:: none
        
        ::
        
            ClassName(
                telegram=<telegram>,
                type=<type>,
                parameter_mode=<parameter_mode>,
                ...
            )
            
        .. highlight:: default
        """
        # str(self.parameter_error) must be used instead of
        # self.parameter_error, because the latter is displayed as just a
        # number instead of its __repr__ or __str__. This is probably caused by
        # the fact that that ParameterError inherits int.
        return type(self).__name__ + ('(\n'
            f'    telegram={self.telegram},\n'
            f'    type={repr(self.type)},\n'
            f'    parameter_mode={repr(self.parameter_mode)},\n'
            f'    parameter_number={self.parameter_number},\n'
            f'    parameter_index={self.parameter_index},\n'
            f'    parameter_value={self.parameter_value},\n'
            f'    parameter_error={str(self.parameter_error)},\n'
            f'    flag_bits={self.flag_bits},\n'
            f'    frequency={self.frequency},\n'
            f'    temperature={self.temperature},\n'
            f'    current={self.current},\n'
            f'    voltage={self.voltage}\n'
            ')'
        )

    @property
    def parameter_mode(self):
        """Return the parameter mode.
        
        This method is the reverse of
        :meth:`TelegramBuilder.set_parameter_mode`: it automatically converts
        a parameter access or response code to a human-readable string.
        
        Raises:
            ValueError: If the parameter code of the telegram is invalid.
        """
        code = self.telegram.parameter_code.value
        return get_parameter_mode(self.type, code)
        
    @property
    def parameter_number(self):
        """Return the parameter number."""
        return self.telegram.parameter_number.value
        
    @property
    def parameter_index(self):
        """Return the parameter index."""
        return self.telegram.parameter_index.value
    
    @property
    def parameter_value(self):
        """Return the parameter value."""
        return self.telegram.parameter_value.value
    
    @property
    def parameter_error(self):
        """Return the parameter error.
        
        Returns:
            A member of the :class:`~turboctl.telegram.codes.ParameterError` 
            enum, or ``None``, if :attr:`parameter_mode` isn't ``'error'``.
            
        Raises:
            ValueError: If the error number isn't valid.
        """
        # The error code could be easily read from parameter_value
        # (which could be changed to always give the value as an Uint if
        # parameter_mode is 'error'). This property mostly exists to aid in
        # debugging; __str__ displays its value, which makes it easy to see
        # which error has occurred without looking up the meaning of the error
        # code.
        # There's no equivalent TelegramBuilder.set_error_code method, because
        # it isn't really needed, and adding that bit of symmetry wouldn't be
        # worth the increased complexity.

        if self.parameter_mode != 'error':
            return None
        
        number = Uint(self.telegram.parameter_value).value
        try:
            return ParameterError(number)
        except KeyError:
            raise ValueError(f'invalid parameter error number: {number}')
    
    @property
    def flag_bits(self):
        """Return the control or status bits as a list of those
        :class:`~turboctl.telegram.codes.StatusBits` or
        :class:`~turboctl.telegram.codes.ControlBits` members that are set to 
        1 in the telegram.
        """
        bits = self.telegram.flag_bits.value
        enum = ControlBits if self.type == 'query' else StatusBits
        return [enum(i) for i, char in enumerate(bits) if char == '1']
    
    @property
    def frequency(self):
        """Return the frequency."""
        return self.telegram.frequency.value
    
    @property
    def temperature(self):
        """Return the temperature."""
        return self.telegram.temperature.value
    
    @property
    def current(self):
        """Return the current."""
        return self.telegram.current.value
    
    @property
    def voltage(self):
        """Return the voltage."""
        return self.telegram.voltage.value
    

def checksum(bytes_: bytes) -> int:    
    """Compute a checksum for a telegram."""
    checksum = 0
    for i in bytes_:
        checksum = checksum ^ i  
    return checksum



"""This module can be used to simulate the I/O behaviour of a TURBOVAC 
pump and thus enable testing the program without access to an actual pump.
"""
import threading

from turboctl.telegram.parser import PARAMETERS
from turboctl.telegram.telegram import TelegramBuilder, TelegramReader
from turboctl.virtualpump.virtualconnection import VirtualConnection
from turboctl.virtualpump.hardware_component import HardwareComponent
from turboctl.virtualpump.parameter_component import (ExtendedParameters,
                                                      ParameterComponent)

class VirtualPump():
    """This class simulates a TURBOVAC pump and tries to respond to 
    signals the same way a physical pump would. This makes it possible
    to test the ``turboctl`` package without connecting to a physical pump.
    
    Attributes:

        connection(:class:`~turboctl.virtualpump.virtualconnection\
.VirtualConnection`):
            Simulates the serial connection.    
        
        parameter_component (:class:`~turboctl.virtualpump\
.parameter_component.ParameterComponent`):
            Handles access to pump parameters.
        
        hardware_component (:class:`~turboctl.virtualpump\
.hardware_component.HardwareComponent`):
            Simulates pump hardware.
    """
    # The first line of the docstring overrides the default signature generated
    # by Sphinx, and thus prevents PARAMETERS from being expanded.    
    
    def __init__(self):
        """Initialize a new :class:`VirtualPump`."""
        
        # The lock prevents two parallel threads from accessing the 
        # same data at the same time.
        self.lock = threading.Lock()
        
        # Custom parameter sets cannot be specified, since HardwareComponent
        # needs certain hardware-related parameters to be present in order to
        # function.
        ext_parameters = ExtendedParameters(PARAMETERS)
        self.connection = VirtualConnection(self.process)
        self.parameter_component = ParameterComponent(ext_parameters)
        self.hardware_component = HardwareComponent(ext_parameters, self.lock)
                        
    def __enter__(self):
        """Called at the beginning of a ``with`` block; returns
        *self*.
        """
        return self
    
    def __exit__(self, type_, value, traceback):
        """Called upon exiting a ``with`` block; calls
        :meth:`stop`.
        """
        self.stop()
                    
    def stop(self):
        """Close parallel threads by calling
        :attr:`connection`:meth:`.close()
        <turboctl.virtualpump.virtualconnection.VirtualConnection.close>`
        and
        :attr:`hardware_component`:meth:`.stop()
        <turboctl.virtualpump.hardware_component.HardwareComponent.stop>`.
        """
        self.connection.close()
        self.hardware_component.stop()
        
    def process(self, bytes_in):
        """Process incoming data.
        
        This function processes the data sent to the virtual pump by
        interpreting it as a telegram and performing any commands 
        specified by it (such as changing parameter values).     
        A reply telegram is then formed and its contents returned.
                
        Args:
            bytes_in: The telegram sent to the pump as a :class:`bytes` object.
            
        Returns:
            The telegram sent back from the pump as a :class:`bytes` object.
            If *bytes_in* cannot be interpreted as a valid telegram,
            the return value will be an empty :class:`bytes` object.
        """
        try:
            builder = TelegramBuilder().from_bytes(bytes_in)
        except ValueError:
            return bytes()
        
        query = TelegramReader(builder.build('query'), 'query')
        
        self.lock.acquire()
        # VirtualConnection runs this function in a parallel thread,
        # so attribute access from other threads is locked while this part is
        # executed in order to prevent race conditions.
        self.parameter_component.handle_parameter(query, builder)
        self.hardware_component.handle_hardware(query, builder)
        self.lock.release()
        
        return bytes(builder.build('reply'))



"""This module defines the :class:`QueueFile` class."""

import queue
import io
import itertools


class QueueFile:
    """This class wraps a :class:`queue.Queue` object in the interface
    of a file-like object.

    The :class:`QueueFile` class makes redirecting input and output
    between different parts of a program easier than it would be using
    standard file objects, such as :class:`io.StringIO` or those
    created by :func:`pty.openpty`.
    :class:`QueueFile` is thread-safe, can be read from and written to
    at the same time, and its blocking behaviour can be easily
    controlled.
    (If there is nothing to read, reading from :class:`io.StringIO`
    never blocks, but reading from an object created with
    :func:`pty.openpty` always does.)

    Attributes:
        queue (queue.Queue):
            The wrapped :class:`~queue.Queue` object.
            Strings are saved in the queue one character at a time.

        block (bool):
            A flag indicating whether :meth:`read` and
            :meth:`readline` should block or not.
    """

    def __init__(self, block=False):
        """Intialize a new :class:`QueueFile` with :attr:`block` set
        to the value given as an argument.
        """
        self.queue = queue.Queue()
        self.block = block

    def write(self, string):
        """Write *string* to the queue and return the number of
        characters written.

        This method emulates :meth:`io.TextIOBase.write`.
        """
        # io.TextIOBase.write() doesn't accept keyword arguments,
        # so we can safely change the argument name from *s* to
        # *string* without breaking anything.
        for char in string:
            self.queue.put(char)
        return len(string)

    def read(self, size=-1):
        """Read and return a string of *size* characters from the queue.

        If the queue contains less than *size* characters or *size* is
        ``None`` or less than ``1``, all characters in the queue are
        read.

        This method emulates :meth:`io.TextIOBase.read`.
        """
        return self._read(size)

    def readline(self, size=-1):
        """Like :meth:`read`, but reading characters is stopped if a
        newline character is encountered. If this happens, the string
        returned includes that newline character.

        This method emulates :meth:`io.TextIOBase.readline`.
        """
        return self._read(size, endchar='\n')

    def _read(self, size=-1, endchar=None):
        """This method provides the functionality shared by
        read() and readline().
        """
        buffer = io.StringIO()

        if not size or size < 0:
            # An infinite generator that yields 0,1,2,...
            iterations = itertools.count()
        else:
            iterations = range(size)

        # Read a finite or an infinite number of characters.
        for _ in iterations:
            try:
                char = self.queue.get(block=self.block)
                buffer.write(char)
            except queue.Empty:
                break

            if endchar and char == endchar:
                break

        return buffer.getvalue()

    def flush(self):
        """This method does nothing, but its existence makes it
        possible to :func:`print` to this object with ``flush=True``
        without raising an :class:`AttributeError`."""